#!/usr/bin/env python

import os
import csv
import sys
import glob
import logging
import argparse
import traceback

import Bio
import numpy as np
import pandas as pd
from tqdm import tqdm
from Bio import SeqIO
import concurrent.futures
from concurrent import futures
from inStrain import SNVprofile
from collections import defaultdict

import inStrain.SNVprofile
import inStrain.controller
import inStrain.profileUtilities

def calculate_gene_metrics(IS, Gdb, gene2sequence, **kwargs):
    '''
    Calculate the metrics of all genes on a parallelized scaffold-level basis
    '''
    # get arguments for the wrapper
    p = int(kwargs.get('processes', 6))

    # figure out your overlap level
    scaffolds_with_genes = set(Gdb['scaffold'].unique())
    scaffolds_in_IS = set(IS._get_covt_keys())
    scaffolds_to_profile = scaffolds_with_genes.intersection(scaffolds_in_IS)
    logging.info("{0} scaffolds with genes, {1} in the IS, {2} to compare".format(
            len(scaffolds_with_genes), len(scaffolds_in_IS), len(scaffolds_to_profile)))
    # print("{0} scaffolds with genes, {1} in the IS, {2} to compare".format(
    #         len(scaffolds_with_genes), len(scaffolds_in_IS), len(scaffolds_to_profile)))

    # iterate
    GeneProfiles = []

    # Make a global scaff table
    global CumulativeSNVtable
    CumulativeSNVtable = IS.get('cumulative_snv_table')
    if len(CumulativeSNVtable) > 0:
        CumulativeSNVtable = CumulativeSNVtable.sort_values('mm')
    else:
        CumulativeSNVtable = pd.DataFrame(columns=['scaffold'])

    if p > 1:
        ex = concurrent.futures.ProcessPoolExecutor(max_workers=p)
        total_cmds = len([x for x in iterate_commands(IS, scaffolds_to_profile, Gdb, gene2sequence, kwargs)])
        wait_for = [ex.submit(profile_genes_wrapper, cmd) for cmd in iterate_commands(IS, scaffolds_to_profile, Gdb, gene2sequence, kwargs)]
        for f in tqdm(futures.as_completed(wait_for), total=total_cmds, desc='Running gene-level calculations on scaffolds'):
            try:
                results = f.result()
                GeneProfiles.append(results)
            except:
                logging.error("We had a failure! Not sure where!")

    else:
        for cmd in tqdm(iterate_commands(IS, scaffolds_to_profile, Gdb, gene2sequence, kwargs),
                        desc='Running gene-level calculations on scaffolds',
                        total = len(scaffolds_to_profile)):
            GeneProfiles.append(profile_genes_wrapper(cmd))

    # Return
    name2result = {}
    for i, name in enumerate(['coverage', 'clonality', 'SNP_density', 'SNP_mutation_types']):
        name2result[name] = pd.concat([G[i] for G in GeneProfiles])

    return name2result

def profile_genes(IS, scaffold, gdb, gene2sequence, **kwargs):
    '''
    This is the money that gets multiprocessed

    Relies on having a global "CumulativeSNVtable"

    * Calculate the clonality, coverage, linkage, and SNV_density for each gene
    * Determine whether each SNP is synynomous or nonsynonymous
    '''
    # Calculate gene-level coverage
    covTs = IS.get('covT', scaffolds=[scaffold])
    if scaffold not in covTs:
        logging.info("{0} isnt in covT!".format(scaffold))
        cdb = pd.DataFrame()
    else:
        covT = covTs[scaffold]
        cdb = calc_gene_coverage(gdb, covT)
        del covT

    # Calculate gene-level clonality
    covTs = IS.get('clonT', scaffolds=[scaffold])
    if (covTs is None):
        logging.info("{0} isnt in clovT!".format(scaffold))
        cldb = pd.DataFrame()
    if scaffold not in covTs:
        logging.info("{0} isnt in clovT!".format(scaffold))
        cldb = pd.DataFrame()
    else:
        covT = covTs[scaffold]
        cldb = calc_gene_clonality(gdb, covT)
        del covT

    # I DECIDED THIS DOESN'T MAKE SENSE TO DO

    # # Calculate gene-level linkage
    # Ldb = IS.get('raw_linkage_table')
    # Ldb = Ldb[Ldb['scaffold'] == 'scaffold'].sort_values('mm')
    # if len(Ldb) == 0:
    #     ldb = pd.DataFrame()
    # else:
    #     ldb = calc_gene_linkage(gdb, ldb)

    # Calculate gene-level SNP densisty
    # Ldb = IS.get('cumulative_snv_table')
    # Ldb = Ldb[Ldb['scaffold'] == scaffold].sort_values('mm')
    # THIS IS A GLOBAL NOW

    Ldb = CumulativeSNVtable[CumulativeSNVtable['scaffold'] == scaffold]

    if len(Ldb) == 0:
        ldb = pd.DataFrame()
    else:
        ldb = calc_gene_snp_density(gdb, Ldb)

    # Determine whether SNPs are synonmous or non-synonmous
    # Sdb = IS.get_nonredundant_snv_table()
    scdb = Ldb
    if (scdb is None) or (len(scdb) == 0):
        Sdb = pd.DataFrame()
        sdb = pd.DataFrame()
    else:
        Sdb = scdb.drop_duplicates(subset=['scaffold', 'position'], keep='last')\
                    .sort_index().drop(columns=['mm'])

        Sdb = Sdb[Sdb['scaffold'] == scaffold]
        Sdb = Sdb[Sdb['cryptic'] == False]

        if 'morphia' in Sdb.columns:
            Sdb['morphia'] = Sdb['morphia'].astype(int)
            # Sdb = Sdb[Sdb['morphia'] == 2]
            Sdb = Sdb[(Sdb['morphia'] > 0) & (Sdb['morphia'] <= 2)]
            #Sdb = Sdb.drop(columns="morphia")

        elif 'allele_count' in Sdb.columns:
            Sdb['allele_count'] = Sdb['allele_count'].astype(int)
            # Sdb = Sdb[Sdb['allele_count'] == 2]
            Sdb = Sdb[(Sdb['allele_count'] > 0) & (Sdb['allele_count'] <= 2)]
            #Sdb = Sdb.drop(columns="allele_count")

        Sdb = Sdb.drop(columns="cryptic")
        Sdb['position'] = Sdb['position'].astype(int)

        if len(Sdb) == 0:
            sdb = pd.DataFrame()
        else:
            sdb = characterize_SNPs(gdb, Sdb, gene2sequence)

    assert len(Sdb) == len(sdb)
    if len(Sdb) > 0:
        sdb = pd.merge(Sdb, sdb, on=['position'], how='left').reset_index(drop=True)

    results = (cdb, cldb, ldb, sdb)

    return results

def calc_gene_coverage(gdb, covT):
    '''
    Gene-level and mm-level coverage
    '''
    table = defaultdict(list)

    # for mm in sorted(list(covT.keys())):
    #     cov = inStrain.profileUtilities._mm_counts_to_counts_shrunk(covT, mm)
    for mm, cov in iterate_covT_mms(covT):
        if len(cov) == 0:
            continue

        for i, row in gdb.iterrows():
            gcov = cov.loc[int(row['start']):int(row['end'])]
            gLen = abs(row['end'] - row['start']) + 1

            table['gene'].append(row['gene'])
            table['coverage'].append(gcov.sum() / gLen)
            table['breadth'].append(len(gcov) / gLen)
            table['mm'].append(mm)

    return pd.DataFrame(table)

def iterate_clonT_mms(clonT):
    p2c = {}
    mms = sorted([int(mm) for mm in list(clonT.keys())])
    for mm in mms:
        for pos, val in clonT[mm].items():
            p2c[pos] = val

        inds = []
        vals = []
        for ind in sorted(p2c.keys()):
            inds.append(ind)
            vals.append(p2c[ind])

        yield mm, pd.Series(data = vals, index = np.array(inds).astype('int'))

def iterate_covT_mms(clonT):
    counts = pd.Series()
    mms = sorted([int(mm) for mm in list(clonT.keys())])
    for mm in mms:
        count = clonT[mm]
        counts = counts.add(count, fill_value=0)
        yield mm, counts
    # p2c = {}
    # mms = sorted([int(mm) for mm in list(clonT.keys())])
    # for mm in mms:
    #     for pos, val in clonT[mm].items():
    #         p2c[pos] = val
    #
    #     inds = []
    #     vals = []
    #     for ind in sorted(p2c.keys()):
    #         inds.append(ind)
    #         vals.append(p2c[ind])
    #
    #     yield mm, pd.Series(data = vals, index = np.array(inds).astype('int'))

def calc_gene_clonality(gdb, clonT):
    '''
    Gene-level and mm-level clonality
    '''
    table = defaultdict(list)

    for mm, cov in iterate_clonT_mms(clonT):
    # for mm in sorted(list(clonT.keys())):
    #     cov = inStrain.profileUtilities._get_basewise_clons3(clonT, mm)
        if len(cov) == 0:
            continue

        for i, row in gdb.iterrows():
            gcov = cov.loc[int(row['start']):int(row['end'])]
            gLen = abs(row['end'] - row['start']) + 1

            table['gene'].append(row['gene'])

            try:
                microdiversity = 1 - gcov.mean()
            except :
                microdiversity = np.nan

            table['clonality'].append(gcov.mean())
            table['microdiversity'].append(microdiversity)
            table['masked_breadth'].append(len(gcov) / gLen)
            table['mm'].append(mm)

    return pd.DataFrame(table)

# def calc_gene_linkage(gdb, ldb, on='r2'):
#     '''
#     Gene-level and mm-level clonality
#     '''
#     table = defaultdict(list)
#
#     for mm in sorted(ldb['mm'].unique()):
#         db = ldb[ldb['mm'] <= mm].drop_duplicates(subset=['scaffold', 'position_A', 'position_B'], keep='last')
#         cov = db.set_index('position_A')[on].sort_index()
#         if len(cov) == 0:
#             continue
#
#         for i, row in gdb.iterrows():
#             gcov = cov.loc[int(row['start']):int(row['end'])]
#             gLen = abs(row['end'] - row['start']) + 1
#
#             table['gene'].append(row['gene'])
#             table['linkage'].append(gcov.mean())
#             table['mm'].append(mm)
#
#     return pd.DataFrame(table)

def calc_gene_snp_density(gdb, ldb):
    '''
    Gene-level and mm-level clonality
    '''
    table = defaultdict(list)

    for mm in sorted(ldb['mm'].unique()):
        db = ldb[ldb['mm'] <= mm].drop_duplicates(subset=['scaffold', 'position'], keep='last')
        cov = db.set_index('position')['refBase'].sort_index()
        if len(cov) == 0:
            continue

        for i, row in gdb.iterrows():
            gcov = cov.loc[int(row['start']):int(row['end'])]
            gLen = abs(row['end'] - row['start']) + 1

            table['gene'].append(row['gene'])
            table['SNPs_per_bp'].append(len(gcov) / gLen)
            table['mm'].append(mm)

    return pd.DataFrame(table)

def characterize_SNPs(gdb, Sdb, gene2sequence):
    '''
    Determine the type of SNP (synonymous, non-synynomous, or intergenic)
    '''
    table = defaultdict(list)
    for i, row in Sdb.iterrows():
        db = gdb[(gdb['start'] <= row['position']) & (gdb['end'] >= row['position'])]
        if len(db) == 0:
            table['position'].append(row['position'])
            table['mutation_type'].append('I')
            table['mutation'].append('')
            table['gene'].append('')
        elif len(db) > 1:
            table['position'].append(row['position'])
            table['mutation_type'].append('M')
            table['mutation'].append('')
            table['gene'].append(','.join(db['gene'].tolist()))
        else:
            # Get the original sequence
            original_sequence = gene2sequence[db['gene'].tolist()[0]]
            if db['direction'].tolist()[0] == '-1':
                original_sequence = original_sequence.reverse_complement()

            # Make the new sequence
            snp_start = row['position'] - db['start'].tolist()[0]
            new_sequence = original_sequence.tomutable()
            new_sequence[snp_start] = row['varBase']
            if new_sequence[snp_start] == original_sequence[snp_start]:
                new_sequence[snp_start] = row['conBase']
            new_sequence = new_sequence.toseq()

            # Translate
            if db['direction'].tolist()[0] == '-1':
                old_aa_sequence = original_sequence.reverse_complement().translate()
                new_aa_sequence = new_sequence.reverse_complement().translate()
            else:
                old_aa_sequence = original_sequence.translate()
                new_aa_sequence = new_sequence.translate()

            # old_aa_sequence = original_sequence.translate()
            # new_aa_sequence = new_sequence.translate()

            # Find mutation
            mut_type = 'S'
            mut = 'S:' + str(snp_start)
            for aa in range(0, len(old_aa_sequence)):
                if new_aa_sequence[aa] != old_aa_sequence[aa]:
                    mut_type = 'N'
                    mut = 'N:' + str(old_aa_sequence[aa]) + str(snp_start) + str(new_aa_sequence[aa])
                    break

            # Store
            table['position'].append(row['position'])
            table['mutation_type'].append(mut_type)
            table['mutation'].append(mut)
            table['gene'].append(db['gene'].tolist()[0])

    return pd.DataFrame(table)

def iterate_commands(IS, scaffolds_to_profile, Gdb, gene2sequence, kwargs):
    '''
    Break into individual scaffolds
    '''
    for scaffold, gdb in Gdb.groupby('scaffold'):
        if scaffold not in scaffolds_to_profile:
            continue

        cmd = Command()
        cmd.IS = IS
        cmd.scaffold = scaffold
        cmd.gdb = gdb
        cmd.arguments = kwargs
        cmd.gene2sequence = {g:gene2sequence[g] for g in gdb['gene'].tolist()}

        yield cmd

def profile_genes_wrapper(cmd):
    '''
    Take a command and profile the scaffold
    '''
    logging.debug('running {0}'.format(cmd.scaffold))
    try:
        return profile_genes(cmd.IS, cmd.scaffold, cmd.gdb, cmd.gene2sequence, **cmd.arguments)
    except Exception as e:
        print(e)
        traceback.print_exc()
        logging.error("whole scaffold exception- {0}".format(str(cmd.scaffold)))
        return pd.DataFrame({'Failed':[True]})

class Command():
    def __init__(self):
        pass

def parse_genes(gene_file, **kwargs):
    if gene_file[-4:] == '.fna':
        return parse_prodigal_genes(gene_file)
    elif ((gene_file[-3:] == '.gb') | (gene_file[-4:] == '.gbk')):
        return parse_genbank_genes(gene_file)
    else:
        print("I dont know how to process {0}".format(gene_file))
        raise Exception

def parse_prodigal_genes(gene_fasta):
    '''
    Parse the prodigal .fna file

    Return a datatable with gene info and a dictionary of gene -> sequence
    '''
    table = defaultdict(list)
    gene2sequence = {}
    for record in SeqIO.parse(gene_fasta, 'fasta'):
        gene = str(record.id)

        table['gene'].append(gene)
        table['scaffold'].append("_".join(gene.split("_")[:-1]))
        table['direction'].append(record.description.split("#")[3].strip())
        table['partial'].append('partial=00' not in record.description)

        # NOTE: PRODIGAL USES A 1-BASED INDEX AND WE USE 0, SO CONVERT TO 0 HERE
        table['start'].append(int(record.description.split("#")[1].strip())-1)
        table['end'].append(int(record.description.split("#")[2].strip())-1)

        gene2sequence[gene] = record.seq

    Gdb = pd.DataFrame(table)
    logging.info("{0}% of the input {1} genes were marked as incomplete".format((len(Gdb[Gdb['partial'] == True])/len(Gdb))*100, len(Gdb)))

    return Gdb, gene2sequence

def parse_genbank_genes(gene_file, gene_name='gene'):
    table = defaultdict(list)
    gene2sequence = {}
    for record in SeqIO.parse(gene_file, 'gb'):
        scaffold = record.id
        for feature in record.features:
            if feature.type == 'CDS':
                gene = feature.qualifiers[gene_name][0]
                loc = feature.location
                if type(loc) is Bio.SeqFeature.CompoundLocation:
                    partial = 'compound'
                else:
                    partial = False

                table['gene'].append(gene)
                table['scaffold'].append(scaffold)
                table['direction'].append(feature.location.strand)
                table['partial'].append(partial)

                table['start'].append(loc.start)
                table['end'].append(loc.end - 1)

                gene2sequence[gene] = feature.location.extract(record).seq

    Gdb = pd.DataFrame(table)
    logging.info("{0}% of the input {1} genes were marked as compound".format((len(Gdb[Gdb['partial'] != False])/len(Gdb))*100, len(Gdb)))

    return Gdb, gene2sequence

def get_gene_info(IS,  ANI_level=0):
    #IS = inStrain.SNVprofile.SNVprofile(IS_loc)

     # Get the mm level
    mm = _get_mm(IS, ANI_level)

    # Load all genes
    Gdb = IS.get('genes_table')

    # Get the coverage
    db = IS.get('genes_coverage')
    db = db[db['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['gene'], keep='last')
    del db['mm']
    Gdb = pd.merge(Gdb, db, on='gene', how='left')

    # Get the clonality
    db = IS.get('genes_clonality')
    db = db[db['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['gene'], keep='last')
    del db['mm']
    Gdb = pd.merge(Gdb, db, on='gene', how='left')

    # Get the SNP density
    db = IS.get('genes_SNP_density')
    if len(db) > 0:
        db = db[db['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['gene'], keep='last')
        del db['mm']
        Gdb = pd.merge(Gdb, db, on='gene', how='left')
    else:
        logging.info('You have no SNPs! Skipping SNP density calculation')

    Gdb['min_ANI'] = ANI_level

    return Gdb

def _get_mm(IS, ANI):
    '''
    Get the mm corresponding to an ANI level in an IS
    '''
    if ANI > 1:
        ANI = ANI / 100

    rLen = IS.get('read_report')['mean_pair_length'].tolist()[0]
    mm = int(round((rLen - (rLen * ANI))))
    return mm

class Controller():

    def main(self, args):
        '''
        The main method when run on the command line
        '''
        # Parse arguments
        #args = self.parse_arguments(sys_args)
        args = self.validate_input(args)

        vargs = vars(args)
        IS = vargs.pop('IS')
        GF = vargs.pop('gene_file')

        # Read the genes file
        logging.debug('Loading genes')
        Gdb, gene2sequence = parse_genes(GF, **vargs)

        # Calculate all your parallelized gene-level stuff
        name2result = calculate_gene_metrics(IS, Gdb, gene2sequence, **vargs)

        # Store information
        IS.store('genes_fileloc', GF, 'value', 'Location of genes .faa file that was used to call genes')
        IS.store('genes_table', Gdb, 'pandas', 'Location of genes in the associated genes_file')
        IS.store('genes_coverage', name2result['coverage'], 'pandas', 'Coverage of individual genes')
        IS.store('genes_clonality', name2result['clonality'], 'pandas', 'Clonality of individual genes')
        IS.store('genes_SNP_density', name2result['SNP_density'], 'pandas', 'SNP density of individual genes')
        IS.store('SNP_mutation_types', name2result['SNP_mutation_types'], 'pandas', 'The mutation types of SNPs')

        if vargs.get('store_everything', False):
            IS.store('gene2sequence', gene2sequence, 'pickle', 'Dicitonary of gene -> nucleotide sequence')

        # Store the output
        out_base = IS.get_location('output') + os.path.basename(IS.get('location')) + '_'
        Gdb = get_gene_info(IS)
        Gdb.to_csv(out_base + 'gene_info.tsv', index=False, sep='\t')

        try:
            name2result['SNP_mutation_types'].to_csv(out_base + 'SNP_mutation_types.tsv', index=False, sep='\t')
        except:
            pass

    def validate_input(self, args):
        '''
        Validate and mess with the arguments a bit
        '''
        # Make sure the IS object is OK
        assert os.path.exists(args.IS)
        args.IS = inStrain.SNVprofile.SNVprofile(args.IS)

        # Set up the logger
        log_loc = args.IS.get_location('log') + 'log.log'
        inStrain.controller.setup_logger(log_loc)

        return args

    def parse_arguments(self, args):
        '''
        Argument parsing
        '''
        parser = argparse.ArgumentParser(description= """
            A script that runs gene-based analyses on inStrain SNV profiles\n

            Required input: a prodigal .fna gene calls file (created with prodigal -d). Going to be opinionated about this and not support alternatives.
            This file is easy to recreate from other / custom data - please look under examples to see how it should be formatted.
            """, formatter_class=argparse.RawTextHelpFormatter)

        # Required positional arguments
        parser.add_argument("-i", '--IS', help="an inStrain object", required=True)
        parser.add_argument("-g", "--gene_file", action="store", required=True, \
            help='Path to prodigal .fna genes file.')
        parser.add_argument("-p", "--processes", action="store", default=6, type=int, \
                            help='Threads to use for multiprocessing')

        # Parse
        if (len(args) == 0 or args[0] == '-h' or args[0] == '--help'):
            parser.print_help()
            sys.exit(0)
        else:
            return parser.parse_args(args)

# if __name__ == '__main__':
#     Controller().main(sys.argv[1:])



#!/usr/bin/env python

import os
import sys
import gzip
import pysam
import logging
import argparse
import numpy as np
import pandas as pd
from tqdm import tqdm
from Bio import SeqIO
from collections import defaultdict
import concurrent.futures
from concurrent import futures
import traceback
import itertools
from collections import ChainMap

class Controller():

    def main(self, args):
        '''
        The main method
        '''
        bam = args.bam
        vargs = vars(args)
        del vargs['bam']

        detailed_report = vargs.get('deatiled_read_report', False)
        generate_sam = vargs.get('generate_sam', False)
        out_folder = vargs.get('output', False)

        # Set up the output folder
        if not os.path.isdir(out_folder):
            os.mkdir(out_folder)

        # Set up .fasta file
        FAdb, s2s = load_fasta(args.fasta)

        # Get the paired reads
        scaffolds = list(FAdb['scaffold'].unique())
        if detailed_report:
            Rdic, RR, dRR = load_paired_reads2(bam, scaffolds, **vargs)
        else:
            Rdic, RR = load_paired_reads2(bam, scaffolds, **vargs)
            dRR = None

        # Make a .sam
        if generate_sam:
            print("The ability to make .sam files is not finished yet; sorry!")

        # Save results
        self.write_results(out_folder, RR, dRR, **vargs)

    def write_results(self, out_folder, RR, dRR, **kwargs):
        '''
        Save the results in a folder for the filter_reads module
        '''
        assert os.path.isdir(out_folder)

        RR_loc = os.path.join(out_folder, 'read_report.csv')
        write_read_report(RR, RR_loc, **kwargs)

        if dRR is not None:
            RR_loc = os.path.join(out_folder, 'deatiled_read_report.csv')
            dRR.to_csv(RR_loc, index=False, sep='\t')

def load_paired_reads2(bam, scaffolds, **kwargs):
    '''
    Load paired reads to be profiled

    You have this method do a lot of things because all of these things take lots of RAM, and you want them all to be cleared as soon as possible

    Return a dictionary of results. Some things that could be in it are:
        pair2infoF: A filtered dictionary of read name to number of mismatches
        RR: A summary read reaport
        RR_detailed: A detailed read report
    '''
    # Parse the kwargs
    detailed_report = kwargs.get('deatiled_read_report', False)

    # Get the pairs
    scaff2pair2info = get_paired_reads_multi2(bam, scaffolds, **kwargs)

    # Handle paired-read filtering
    priority_reads_loc = kwargs.get('priority_reads', None)
    priority_reads = load_priority_reads(priority_reads_loc)
    pair2info = paired_read_filter(scaff2pair2info, priority_reads_set=priority_reads, **kwargs)

    # Make read report
    logging.info('Making read report')
    RR = makeFilterReport2(scaff2pair2info, pairTOinfo=pair2info, priority_reads_set=priority_reads, **kwargs)

    # Filter the dictionary
    logging.info('Filtering reads')
    pair2infoF = filter_paired_reads_dict2(pair2info,
        **kwargs)

    if detailed_report:
        dRR = make_detailed_read_report(scaff2pair2info, pairTOinfo=pair2info, version=2)
        return pair2infoF, RR, dRR

    else:
        return pair2infoF, RR

def load_priority_reads(file_loc):
    '''
    Loads a file of reads and returns a set of their names
    '''
    # is it None?
    if file_loc is None:
        return set()

    # Is it zipped?
    if file_loc[-3:] == '.gz':
        reader = gzip.open(file_loc, 'rt')
    else:
        reader = open(file_loc, 'r')

    # Figure out the type
    for line in reader.readlines():
        if line[0] == '@':
            TYPE = 'fastq'
        else:
            TYPE = 'list'
        break

    reader.close()
    if file_loc[-3:] == '.gz':
        reader = gzip.open(file_loc, 'rt')
    else:
        reader = open(file_loc, 'r')

    reads = set()
    if TYPE == 'fastq':
        for line in reader.readlines():
            if line[0] != '@':
                continue
            reads.add(line[1:].strip())

    elif TYPE == 'list':
        for line in reader.readlines():
            reads.add(line.strip())

    reader.close()

    return reads

def paired_read_filter(scaff2pair2info, priority_reads_set=set(), **kwargs):
    '''
    Filter scaff2pair2info to keep / remove paired / unpaired reads
    '''
    assert type(kwargs.get('priority_reads', 'none')) != type(set())
    priority_reads = priority_reads_set
    pairing_filter = kwargs.get('pairing_filter', 'paired_only')
    pair2info = {}
    assert type(priority_reads) == type(set()), type(priority_reads)

    if pairing_filter == 'paired_only':
        for scaff, p2i in scaff2pair2info.items():
            for p, i in p2i.items():
                if ((i[4] == 2) | (p in priority_reads)):
                    pair2info[p] = i

    elif pairing_filter == 'non_discordant':
        for scaff, p2i in scaff2pair2info.items():
            for p, i in p2i.items():
                # Add it if it's not already in there
                if ((p not in pair2info) | (p in priority_reads)):
                    pair2info[p] = i

                # If it is already in there, that means its concordant, so delete it
                else:
                    del pair2info[p]

    elif pairing_filter == 'all_reads':
        for scaff, p2i in scaff2pair2info.items():
            for p, i in p2i.items():
                if p in pair2info:
                    pair2info[p] = _merge_info(i, pair2info[p])
                else:
                    pair2info[p] = i

    else:
        logging.error("Do not know paired read filter \"{0}\"; crashing now".format(pairing_filter))
        raise Exception

    return pair2info

def _merge_info(i1, i2):
    #{'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
    return np.array([i1[0] + i2[0],
            -2,
            max([i1[2] + i2[2]]),
            i1[3] + i2[3],
            i1[4] + i2[4],
            -1,
            -1], dtype="int64")

def make_detailed_read_report(scaff2pair2info, pairTOinfo=dict(), version=2):
    '''
    Make a detailed pandas dataframe from pair2info
    '''
    if version == 2:
        i2o = {'mm':0, 'insert_dist':1, 'mapq':2, 'length':3, 'reads':4,
                'start':5, 'stop':6}
    elif version == 1:
        i2o = {'mm':0, 'insert_dist':1, 'mapq':2, 'length':3,}

    keepers = pairTOinfo.keys()
    report_keepers =  (len(keepers) > 0)

    table = defaultdict(list)
    for scaff, pair2info in scaff2pair2info.items():
        for pair, array in pair2info.items():
            table['read_pair'].append(pair)
            table['scaffold'].append(scaff)
            if report_keepers:
                table['pass_filters'].append(pair in keepers)

            for item, location in i2o.items():
                table[item].append(array[location])

    return pd.DataFrame(table)

def load_fasta(fasta_file):
    '''
    Load the sequences to be profiled

    Return a table listing scaffold name, start, end
    '''
    # PROFILE ALL SCAFFOLDS IN THE .FASTA FILE
    scaff2sequence = SeqIO.to_dict(SeqIO.parse(fasta_file, "fasta")) # set up .fasta file
    s2l = {s:len(scaff2sequence[s]) for s in list(scaff2sequence.keys())} # Get scaffold2length
    Fdb = pd.DataFrame(list(s2l.items()), columns=['scaffold', 'end'])
    Fdb['start'] = 0

    return Fdb, scaff2sequence # also return s2l - alexcc 5/9/2019: Nah, make it scaff2sequence (s2s) (M.O. 6/10/19)

def filter_paired_reads_dict2(pair2info, **kwargs):
    '''
    Filter the dictionary of paired reads, end with read -> mm
    '''
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Get kwargs
    filter_cutoff = kwargs.get('filter_cutoff', 0.97)
    max_insert_relative = kwargs.get('max_insert_relative', 3)
    min_insert = kwargs.get('min_insert', 50)
    min_mapq = kwargs.get('min_mapq', 2)

    # Get max insert
    max_insert = np.median([value[1] for key, value in pair2info.items() if value[i2o['reads']] == 2]) * max_insert_relative

    # Return dictionary of pairs
    return {key:value[0] for key, value in pair2info.items() if _evaluate_pair2(value,
            filter_cutoff=filter_cutoff,
            max_insert=max_insert,
            min_insert=min_insert,
            min_mapq=min_mapq)}

def makeFilterReport2(scaff2pair2info, pairTOinfo=False, priority_reads_set=set(), **kwargs):
    '''
    Make a scaffold-level report on when reads are filtered using get_paired_reads_multi2

    If you've already filtered out pairs as you'd like, pass in pairTOinfo
    '''
    assert type(kwargs.get('priority_reads', 'none')) != type(set())
    priority_reads = priority_reads_set

    #item2order
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Calculate max insert
    max_insert_relative = kwargs.get('max_insert_relative', 3)
    median_insert = np.median([value[i2o['insert_distance']] for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if value[i2o['reads']] == 2])
    max_insert = median_insert * max_insert_relative

    # Get values
    values = {}
    values['filter_cutoff'] = kwargs.get('filter_cutoff', 0.97)
    values['max_insert'] = max_insert
    values['min_insert'] = kwargs.get('min_insert', 50)
    values['min_mapq'] = kwargs.get('min_mapq', 2)

    # Make report on all scaffolds
    logging.debug('running on all reads')
    table = defaultdict(list)
    table['scaffold'].append('all_scaffolds')
    table['unfiltered_reads'].append(sum([value[i2o['reads']] for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items()]))
    table['unfiltered_pairs'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if value[i2o['reads']] == 2]))
    table['unfiltered_singletons'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if (value[i2o['reads']] == 1)]))
    table['unfiltered_priority_reads'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, value in pair2info.items() if (pair in priority_reads)]))

    if pairTOinfo != False:
        keepers = set(pairTOinfo.keys())
        infos = [info for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items() if pair in keepers]
        table['pass_pairing_filter'].append(len(infos))
    else:
        infos = [info for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items()]

    for att, v in values.items():
        kwargs={att:v}
        table['pass_' + att].append(len([True for info in infos if (_evaluate_pair2(info, **kwargs))]))
    table['filtered_pairs'].append(len([True for info in infos if (_evaluate_pair2(info, **values))]))
    table['filtered_singletons'].append(len([True for info in infos if ((info[i2o['reads']] == 1) & (_evaluate_pair2(info, **values)))]))
    table['filtered_priority_reads'].append(len([True for scaff, pair2info in scaff2pair2info.items() for pair, info in pair2info.items() if ((pair in priority_reads) & (_evaluate_pair2(info, **values)))]))

    for i, att in enumerate(['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']):
        table['mean_' + att].append(np.mean([info[i] for info in infos]))
    table['median_insert'].append(np.median([info[i2o['insert_distance']] for info in infos]))
    table['mean_PID'].append(np.mean([(1 - (float(info[i2o['nm']]) / float(info[i2o['length']]))) for info in infos]))

    logging.debug('running on individual scaffolds')
    for scaff, pair2info in scaff2pair2info.items():
        table['scaffold'].append(scaff)
        table['unfiltered_reads'].append(sum([value[i2o['reads']] for pair, value in pair2info.items()]))
        table['unfiltered_pairs'].append(len([True for pair, value in pair2info.items() if value[i2o['reads']] == 2]))
        table['unfiltered_singletons'].append(len([True for pair, info in pair2info.items() if (info[i2o['reads']] == 1)]))
        table['unfiltered_priority_reads'].append(len([True for pair, info in pair2info.items() if (pair in priority_reads)]))

        if pairTOinfo != False:
            keepers = set(pairTOinfo.keys())
            infos = [info for pair, info in pair2info.items() if pair in keepers]
            table['pass_pairing_filter'].append(len(infos))
        else:
            infos = [info for pair, info in pair2info.items()]

        for att, v in values.items():
            kwargs={att:v}
            table['pass_' + att].append(len([True for info in infos if (_evaluate_pair2(info, **kwargs))]))
        table['filtered_pairs'].append(len([True for info in infos if (_evaluate_pair2(info, **values))]))
        table['filtered_singletons'].append(len([True for info in infos if ((info[i2o['reads']] == 1) & (_evaluate_pair2(info, **values)))]))
        table['filtered_priority_reads'].append(len([True for pair, info in pair2info.items() if ((pair in priority_reads) & (_evaluate_pair2(info, **values)))]))

        for i, att in enumerate(['mistmaches', 'insert_distance', 'mapq_score', 'pair_length']):
            table['mean_' + att].append(np.mean([info[i] for pair, info in pair2info.items()]))
        table['median_insert'].append(np.median([value[1] for key, value in pair2info.items()]))
        table['mean_PID'].append(np.mean([(1 - (float(info[i2o['nm']]) / float(info[i2o['length']]))) for pair, info in pair2info.items()]))

    return pd.DataFrame(table)

def write_read_report(RR, location, **kwargs):
    # Get header materials
    values = {}
    values['filter_cutoff'] = kwargs.get('filter_cutoff', 0.97)
    values['max_insert_relative'] = kwargs.get('max_insert_relative', 3)
    values['min_insert'] = kwargs.get('min_insert', 50)
    values['min_mapq'] = kwargs.get('min_mapq', 2)
    values['pairing_filter'] = kwargs.get('pairing_filter', 'paired_only')

    # Write header
    os.remove(location) if os.path.exists(location) else None
    f = open(location, 'a')
    f.write("# {0}\n".format(' '.join(["{0}:{1}".format(k, v) for k, v in values.items()])))

    # Write csv
    RR.to_csv(f, index=False, sep='\t')

    # Close
    f.close()

i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}
def _evaluate_pair2(value, max_insert=1000000, filter_cutoff=-1, min_insert=-1,
                        min_mapq=-1):
    # calculate PID for this pair
    PID = 1 - (float(value[i2o['nm']]) / float(value[i2o['length']]))

    # If this is a pair:
    if ((value[i2o['reads']] == 2) & (value[i2o['insert_distance']] != -1)):
        # See if it passes filtering
        if ((PID > filter_cutoff) & (value[i2o['insert_distance']] > min_insert) & (value[i2o['insert_distance']] < max_insert)\
            & (value[i2o['mapq']] > min_mapq)):
            return True
        else:
            return False

    # If this is not a pair
    else:
        if ((PID > filter_cutoff) & (value[i2o['mapq']] > min_mapq)):
            return True
        else:
            return False

def get_paired_reads_multi2(bam, scaffolds, **kwargs):
    '''
    Returns scaffold 2 read 2 info
    '''
    # Initialize dictionary
    dicts = []
    p = int(kwargs.get('processes', 6))
    ret_total = kwargs.get('ret_total', False)

    # Do the multiprocessing
    if p > 1:
        executor = concurrent.futures.ProcessPoolExecutor(max_workers=p)

        total_cmds = len([x for x in iterate_read_commands(scaffolds, bam, kwargs)])

        wait_for = [executor.submit(scaffold_profile_wrapper2, cmd) for cmd in iterate_read_commands(scaffolds, bam, kwargs)]

        for f in tqdm(futures.as_completed(wait_for), total=total_cmds, desc='Getting read pairs: '):
            try:
                results = f.result()
                dicts.append(results)
            except:
                logging.error("We had a failure! Not sure where!")

    else:
        for cmd in tqdm(iterate_read_commands(scaffolds, bam, kwargs), desc='Getting read pairs: ', total=len(scaffolds)):
            dicts.append(scaffold_profile_wrapper2(cmd))

    # Try and save any ones that failed
    failed_scaffs = set(scaffolds) - set([dict[0] for dict in dicts])
    if len(failed_scaffs) > 0:
        logging.error("The following scaffolds failed- I'll try again {0}".format('\n'.join(failed_scaffs)))
        for cmd in tqdm(iterate_read_commands(failed_scaffs, bam, kwargs), desc='Getting read pairs for previously failed scaffolds: ', total=len(failed_scaffs)):
            dicts.append(scaffold_profile_wrapper2(cmd))

    allPair2info = {}
    for s, pair2info in dicts:
        if pair2info != False:
            allPair2info[s] = {}
            for k, v in pair2info.items():
                allPair2info[s][k] = v

    return allPair2info

class read_command():
    '''
    This class just holds all of the mumbo-jumbo needed to profile a scaffold
    '''
    def __init__(self):
        pass

def iterate_read_commands(scaffolds, bam, profArgs):
    '''
    Make and iterate profiling commands
    Doing it in this way makes it use way less RAM
    '''
    for scaff in scaffolds:
        yield [scaff, bam]


def scaffold_profile_wrapper2(cmd):
    '''
    Take a scaffold and get the reads

    cmd[0] = scaffold, cmd[1] = bam, cmd[2] = kwargs
    '''
    logging.debug('running {0} reads'.format(cmd[0]))
    try:
        pair2info =  get_paired_reads2(cmd[1], cmd[0])
        logging.debug('returning {0} reads'.format(cmd[0]))
        return cmd[0], pair2info
    except Exception as e:
        print(e)
        traceback.print_exc()
        logging.error("whole scaffold exception- {0}".format(str(cmd[0])))
        return cmd[0], False

def get_paired_reads2(bam, scaff):
    '''
    Filter reads from a .bam file

    As opposed to get_paired_reads, this gets all reads and lets you filter out pairs (or not) later. This just means adding a bit to pair2info

    Returns:
        pair2info: dictionary of read pair -> (mismatches, insert distance (-1 if number of reads is not 2), mapq score (highest), combined length, number of reads,
                                                reference_position_0, reference_position_1 (only used for small things))
    '''
    # item2order, to make things more readable
    i2o = {'nm':0, 'insert_distance':1, 'mapq':2, 'length':3, 'reads':4, 'start':5, 'stop':6}

    # Initialize
    pair2info = {} # Information on pairs
    samfile = pysam.AlignmentFile(bam)

    try:
        iter = samfile.fetch(scaff)
    except ValueError:
        logging.error("{0} is not in .bam file".format(scaff))
        return {}

    for read in iter:
        # Dont use unmapped reads
        if read.get_reference_positions() == []:
            continue

        # Add a read the first time you see it
        elif read.query_name not in pair2info:
            pair2info[read.query_name] = np.array([read.get_tag('NM'),
                                            -1,
                                            read.mapping_quality,
                                            read.infer_query_length(),
                                            1,
                                            read.get_reference_positions()[0],
                                            read.get_reference_positions()[-1]], dtype="int64")

        # If we've seen this read's pair before
        elif read.query_name in pair2info:
            # Adjust the number of mismatches
            pair2info[read.query_name][i2o['nm']] = int(pair2info[read.query_name][i2o['nm']]) + int(read.get_tag('NM'))

            # Adjust the number of reads
            pair2info[read.query_name][i2o['reads']] += 1

            # Adjust the combined length
            pair2info[read.query_name][i2o['length']] += read.infer_query_length()

            # Adjust the mapQ
            pair2info[read.query_name][i2o['mapq']] = max(pair2info[read.query_name][i2o['mapq']], read.mapping_quality)

            # If the number of reads is 2, calculate insert size
            if pair2info[read.query_name][i2o['reads']] == 2:
                if read.get_reference_positions()[-1] > pair2info[read.query_name][i2o['start']]:
                    pair2info[read.query_name][i2o['insert_distance']] = read.get_reference_positions()[-1] - pair2info[read.query_name][i2o['start']]
                else:
                    pair2info[read.query_name][i2o['insert_distance']] = pair2info[read.query_name][i2o['stop']] - read.get_reference_positions()[0]

            # Otherwise get rid of insert distance
            else:
                pair2info[read.query_name][i2o['insert_distance']] = -1

            # Get rid of start and stop; not needed anymore
            pair2info[read.query_name][i2o['start']] = 0
            pair2info[read.query_name][i2o['stop']] = 0

    return pair2info

def read_report_wrapper(args, FAdb):
    # Get paired reads
    scaffolds = list(FAdb['scaffold'].tolist())
    scaff2pair2info, scaff2total = get_paired_reads_multi(args.bam, scaffolds, processes=args.processes, ret_total=True)

    # Make report
    RR = makeFilterReport(scaff2pair2info, scaff2total, **vars(args))

    # Save report
    write_read_report(RR, args.read_report, **vars(args))



'''
This houses functions that are from an old version of inStrain; useful for testing
'''
import os
import pysam
import pickle
import logging
import itertools
import numpy as np
import pandas as pd
import networkx as nx
from Bio import SeqIO
from tqdm import tqdm
from collections import defaultdict

import inStrain.controller
import inStrain.profileUtilities

# THIS IS CC'S VERSION
def main(args):
    '''
    Main entry point
    '''

    # Set up Logger
    if not args.output:
        log = args.fasta.split("/")[-1].split(".")[0] + '.log'
    else:
        log = args.output + '.log'
    inStrain.controller.setup_logger(log)

    if args.filter_cutoff:
        if float(args.filter_cutoff) < 1:
            strain_pipeline(args, float(args.filter_cutoff))
        else:
            logging.error("Error: Strain level should be a number between 0 and 1.")
            sys.exit(1)
    else:
        strain_pipeline(args, 0.96)

def strain_pipeline(args, filter_cutoff):
    strains = SNVdata()

    # if args.testing:
    #     strains.testing = True

    if not args.output:
        strains.output = args.fasta.split("/")[-1].split(".")[0] + "_" + str(filter_cutoff)
    else:
        strains.output = args.output + "_" + str(filter_cutoff)

    logging.info("Running at resolution: (>" + str(filter_cutoff) + "%)")

    args.genes = None
    strains.get_scaffold_positions(args.genes, args.fasta)
    strains.run_strain_profiler(args.bam, min_cov = int(args.min_cov), min_freq=float(args.min_freq), filter_cutoff = filter_cutoff)

    logging.info("\nCalculating linkage network...\n")
    strains.calc_linkage_network()

    logging.info("Calculating LD...\n")
    strains.calc_ld_all_sites(int(args.min_snp))
    strains.save()

class SNVdata:
    # The class "constructor" - It's actually an initializer
    def __init__(self):

        # Parameters
        self.fasta = ""
        self.bam = ""
        self.results = False
        self.output = None
        self.testing = False
        self.positions = []
        self.fasta_length = 0


        # Data structures
        self.snv_table = None         #
        self.read_to_snvs = None      #
        self.windows_to_snvs = None   #
        self.all_counts = None        # All counts of AGCT for each position
        self.snv_counts = None        # Counts of AGCT for each SNV position
        self.clonality_table = None         # Dict of clonality histograms (lists) by window
        self.snv_graph = None         # Weighted networkx graph that tracks SNVs( 100:A, 100:T are diff nodes) that are linked
        self.position_graph = None    # Unweighted networkx graph that tracks positions that are linked
        self.r2linkage_table = None         # Dict of r2 histograms (list of lists) by window

        # General statistics
        self.coverages = None
        self.total_positions = None
        self.total_snv_sites = None
        self.alpha_snvs = None

    def save(self, size=0):
        if self.results:
            # Generate tables
            if size == 0:
                self.read_to_snvs = None

            logging.info("printing to tables.")
            sample = self.bam.split("/")[-1].split(".bam")[0]

            genome = self.output
            logging.info(genome)
            self.snv_table.to_csv(genome + ".freq",sep='\t')#, quoting=csv.QUOTE_NONE)
            self.clonality_table.to_csv(genome + ".clonal",sep='\t')#, quoting=csv.QUOTE_NONE)
            self.r2linkage_table.to_csv(genome + ".linkage",sep='\t')#, quoting=csv.QUOTE_NONE)
            self.read_report.to_csv(genome + ".readFiltering",sep='\t')#', quoting=csv.QUOTE_NONE)

            f = open(genome + ".data", 'wb')
            pickle.dump(self.__dict__, f, 2)
            f.close()
        else:
            logging.info("No data to save.")

    def load(self, name):
        self.__dict__.clear()
        f = open(name + ".data", 'rb')
        tmp_dict = pickle.load(f)
        f.close()

        self.__dict__.update(tmp_dict)


    def get_scaffold_positions(self, gene_list = None, fasta_file = None):
        ''' Returns a list of windows to record SNVs in'''
        if not fasta_file and not gene_list:
#            logging.info("ERROR: REQUIRED TO SUPPLY FASTA or GENE FILE")
            sys.exit(1)

        if gene_list:
            f = open(gene_list)
            for line in f.readlines():
                self.positions.append([line.split(",")[1],int(line.split(",")[2]),int(line.split(",")[3])])
            f.close()
        else:
            for rec in SeqIO.parse(fasta_file, "fasta"):
                self.fasta_length += len(rec.seq)
                start = 0
                while True:
                    chunk = start + 15000
                    if chunk <= len(rec.seq):
                        self.positions.append([str(rec.id),start,start+10000])
                    else:
                        self.positions.append([str(rec.id),start,len(rec.seq)])
                        break
                    start += 10000
                    start += 1

        self.fasta = fasta_file

    def get_snps_per_gene(self, gene_file = None):
        pass

    def calc_linkage_network(self):
        ''' Calculates the SNV linkage network - saves it as a dictionary of edges in self.snv_net.
        Writes it out to a file, genome.net for reading in through other programs like node2vec
        '''
        G=nx.Graph() # graph between SNPs  (100:A and 100:C are different nodes)
        G_pos = nx.Graph() # graph between positions; unweighted

        logging.info("Calculating SNV linkage network...")
        for read in self.read_to_snvs:
            snvs = self.read_to_snvs[read]
            for pair in itertools.combinations(snvs, 2):
                if G.has_edge(pair[0], pair[1]):
                    G[pair[0]][pair[1]]['weight'] += 1
                else:
                    G.add_edge(pair[0], pair[1], weight = 1)
                    if not G_pos.has_edge(pair[0].split(":")[0], pair[1].split(":")[0]):
                        G_pos.add_edge(pair[0].split(":")[0], pair[1].split(":")[0])

        logging.info(str("Of " + str(self.total_snv_sites) + " SNP-sites, there were " + str(len(G_pos)) + " SNPs that could be linked to at least one other SNP."))
        logging.info("The average SNP was linked to " + str(int(np.mean([int(x[1]) for x in list(G_pos.degree())]))) + " other SNPs.")
        self.snv_graph = G
        self.position_graph = G_pos


    def calc_ld(self, snp_a, snp_b, min_snp, report=False):
        '''
        A function that calculates the LD between two SNPs
        '''

        #distance between snps
        distance = abs(int(snp_a.split("_")[-1]) - int(snp_b.split("_")[-1]))

        #calculate allele frequencies
        allele_A = major_allele(snp_a,self.snv_counts[snp_a]) #get major allele
        allele_a = minor_allele(snp_a,self.snv_counts[snp_a]) # get minor allele
        freq_A = float(allele_A[1]) / (allele_A[1] + allele_a[1])
        freq_a = float(allele_a[1]) / (allele_A[1] + allele_a[1])

        # get snp B frequencies
        allele_B = major_allele(snp_b,self.snv_counts[snp_b])
        allele_b = minor_allele(snp_b,self.snv_counts[snp_b])
        freq_B = float(allele_B[1]) / (allele_B[1] + allele_b[1])
        freq_b = float(allele_b[1]) / (allele_B[1] + allele_b[1])

        # Get frequencies of linkages
        countAB, countAb, countaB, countab  = 0,0,0,0
        if self.snv_graph.has_edge(allele_A[0], allele_B[0]):
            countAB = self.snv_graph[allele_A[0]][allele_B[0]]['weight']
        if self.snv_graph.has_edge(allele_A[0], allele_b[0]):
            countAb = self.snv_graph[allele_A[0]][allele_b[0]]['weight']
        if self.snv_graph.has_edge(allele_a[0], allele_B[0]):
            countaB = self.snv_graph[allele_a[0]][allele_B[0]]['weight']
        if self.snv_graph.has_edge(allele_a[0], allele_b[0]):
            countab = self.snv_graph[allele_a[0]][allele_b[0]]['weight']

        total = countAB + countAb + countaB + countab

        if report:
            print("total {0}".format(total))
            print([countAB, countAb, countaB, countab])

        #Requires at least min_snp
        if total > min_snp:

            linkage_points_x = []
            linkage_points_y = []
            for point in range(0,countAB):
                linkage_points_x.append(1)
                linkage_points_y.append(1)
            for point in range(0,countAb):
                linkage_points_x.append(1)
                linkage_points_y.append(0)
            for point in range(0,countaB):
                linkage_points_x.append(0)
                linkage_points_y.append(1)
            for point in range(0,countab):
                linkage_points_x.append(0)
                linkage_points_y.append(0)

            freq_AB = float(countAB) / total
            freq_Ab = float(countAb) / total
            freq_aB = float(countaB) / total
            freq_ab = float(countab) / total

            freq_A = freq_AB + freq_Ab
            freq_a = freq_ab + freq_aB
            freq_B = freq_AB + freq_aB
            freq_b = freq_ab + freq_Ab

            linkD = freq_AB - freq_A * freq_B

            if freq_a == 0 or freq_A == 0 or freq_B == 0 or freq_b == 0:
                r2_book = np.nan
            else:
                r2_book = linkD*linkD / (freq_A * freq_a * freq_B * freq_b)


            linkd = freq_ab - freq_a * freq_b
            # calc D-prime
            if (linkd < 0):
                denom = max([(-freq_A*freq_B),(-freq_a*freq_b)])
                d_prime = linkd / denom

            elif (linkD > 0):
                denom = min([(freq_A*freq_b), (freq_a*freq_B)])
                d_prime = linkd / denom
            else:
                d_prime = 0




            # r2 = stats.pearsonr(linkage_points_x, linkage_points_y)[0]
            # r2 = r2 * r2
            # logging.info(r2)
            # logging.info(r2_book)
            r2 = r2_book
            return([distance, r2, linkD, linkd, r2_book, total, countAB, countAb, countaB, countab, allele_A, allele_a, allele_B, allele_b])
            # else:
            #     logging.info("nan")
            # else:
        else:
            return False

    def calc_ld_all_sites(self, min_snp):
        '''
        Calculates Linkage Disequilibrium for all SNVs in a window.
        '''
        r2_total = {}
        sigma_total = {}
        logging.info("Calculating Linkage Disequilibrium")

        for window in self.positions:
            r2_spectrum = []
            sigma_spectrum = []

            window_name = window[0] + ":" + str(window[1]) + ":" + str(window[2])
            window_snvs = self.windows_to_snvs[window_name]

            for edge in self.position_graph.edges(window_snvs):
                snp_a = edge[0]
                snp_b = edge[1]

                ld_result = self.calc_ld(snp_a, snp_b, min_snp)
                if ld_result:
                    r2_spectrum.append(ld_result)

            r2_total[window_name] = r2_spectrum

        # Create r2 linkage table
        r2linkage_table = defaultdict(list)

        for window in r2_total:
            for datum in r2_total[window]:
                r2linkage_table['Window'].append(window)
                r2linkage_table['Distance'].append(datum[0])
                r2linkage_table['r2'].append(datum[1])
                r2linkage_table['linkD'].append(datum[2])
                # r2linkage_table['linkd'].append(datum[3])
                # r2linkage_table['r2_book'].append(datum[4])
                r2linkage_table['total'].append(datum[5])
                r2linkage_table['count_AB'].append(datum[6])
                r2linkage_table['count_Ab'].append(datum[7])
                r2linkage_table['count_aB'].append(datum[8])
                r2linkage_table['count_ab'].append(datum[9])

                r2linkage_table['total_A'].append(datum[10])
                r2linkage_table['total_a'].append(datum[11])
                r2linkage_table['total_B'].append(datum[12])
                r2linkage_table['total_b'].append(datum[13])


        self.r2linkage_table = pd.DataFrame(r2linkage_table)


    def run_strain_profiler(self, bam, min_cov = 5, min_freq = 0.05, filter_cutoff = 0):
        '''
        Main class for finding SNVs and generating data profile for a genome.
        '''
        minimum_mapq = 2
        P2C = {'A':0, 'C':1, 'T':2, 'G':3}
        C2P = {0:'A', 1:'C', 2:'T', 3:'G'}

        #Set up variables
        alpha_snvs = 0
        total_positions = 0
        total_snv_sites = 0

        raw_counts_data = defaultdict(dict) # Set up SNP table
        read_to_snvs = defaultdict(list) # read name -> position : variant base
        snvs_frequencies = defaultdict(int)
        clonality_by_window = defaultdict(list) # window -> [position, clonality]
        windows_to_snvs = defaultdict(list) # window -> position

        all_counts = {} # position -> coverage
        snv_counts = {} # position -> [A, C, T, G]
        coverages = {} # position -> coverage

        # read null model
        null_loc = os.path.dirname(__file__) + '/helper_files/NullModel.txt'
        null_snp_model = inStrain.profileUtilities.generate_snp_model(null_loc)

        ## Start reading BAM
        samfile = pysam.AlignmentFile(bam)
        sample = bam.split("/")[-1].split(".bam")[0]

        if self.testing:
            self.positions = self.positions[0:10]

        # Call read filtering function
        subset_reads, Rdb = inStrain.deprecated_filter_reads.filter_reads(bam, self.positions, self.fasta_length, filter_cutoff, 3, 50, 2)

        # Save the read filtering report
        self.read_report = Rdb

        ### START SNP FINDING

        ## Start looping through each region
        for gene in tqdm(self.positions, desc='Finding SNVs ...'):
            scaff = gene[0]
            window = gene[0] + ":" + str(gene[1]) + ":" + str(gene[2])
            for pileupcolumn in samfile.pileup(scaff, truncate = True, max_depth=100000, stepper = 'nofilter', compute_baq= True, ignore_orphans = True, ignore_overlaps = True,  min_base_quality = 30):
            #for pileupcolumn in samfile.pileup(scaff, gene[1], gene[2], truncate = True, max_depth=100000, stepper = 'nofilter', compute_baq= True, ignore_orphans = True, ignore_overlaps = True,  min_base_quality = 30):
                ## Step 1: Are there any reads at this position?

                position = scaff + "_" + str(pileupcolumn.pos)
                counts = _get_base_counts(pileupcolumn, filtered_reads = subset_reads)

                consensus = False
                # Yes there were reads at this position
                if counts:
                    all_counts[position] = counts

                    if sum(counts) >= min_cov:
                        total_positions += 1
                        pos_clonality = inStrain.profileUtilities.calculate_clonality(counts)
                        clonality_by_window[window].append([position, pos_clonality])
                        consensus = call_snv_site_old(counts, null_snp_model, min_cov = min_cov, min_freq = min_freq)
                        coverages[position] = sum(counts)

                # SEARCH HERE FOR DEBUG
                # if position == "N5_271_010G1_scaffold_114_990":
                #     print(counts)
                #     #counts2 = _get_base_counts_(pileupcolumn, filtered_reads = subset_reads)
                #     reads = []
                #     for pileupread in pileupcolumn.pileups:
                #         if not pileupread.is_del and not pileupread.is_refskip:
                #             if pileupread.alignment.query_name in subset_reads:
                #                 read_name = pileupread.alignment.query_name
                #                 reads.append((read_name, pileupread.alignment.query_sequence[pileupread.query_position]))
                #
                #     print(len(reads))
                #     print('\n'.join([str(r) for r in reads]))
                #     print(consensus)

                ## Strep 2: Is there an SNV at this position?
                if consensus:
                    #min_snp for this site
                    local_min_snp = null_snp_model[sum(counts)]
                    #there's an SNV at this site
                    total_snv_sites += 1
                    windows_to_snvs[window].append(position)
                    snv_counts[position] = counts

                    # Get reads to snvs
                    for pileupread in pileupcolumn.pileups:
                        read_name = pileupread.alignment.query_name
                        if not pileupread.is_del and not pileupread.is_refskip:
                            if read_name in subset_reads:
                                try:
                                    val = pileupread.alignment.query_sequence[pileupread.query_position]
                                    #if value is not the consensus value
                                    if counts[P2C[val]] >= local_min_snp and float(counts[P2C[val]]) / sum(counts) >= min_freq:
                                        #this is a variant read!
                                        read_to_snvs[read_name].append(position + ":" + val)
                                        if val != consensus:
                                            alpha_snvs += 1
                                except KeyError: # This would be like an N or something not A/C/T/G
                                    pass

                    # Add to frequencies
                    nucl_count = 0
                    for nucl in counts:
                        if nucl >= local_min_snp:
                            freq = float(nucl) / float(sum(counts))
                            if freq >= min_freq:
                                snp = position + ":" + C2P[nucl_count]
                                snvs_frequencies[snp] = [freq, window]

                        nucl_count += 1

        #### SNP FINDING COMPLETE
        # Create SNV frequency table
        snv_table = defaultdict(list)
        for snv in snvs_frequencies:
            snv_table['SNV'].append(snv)
            snv_table['freq'].append(snvs_frequencies[snv][0])
            snv_table['Window'].append(snvs_frequencies[snv][1])
        snv_table = pd.DataFrame(snv_table)

        # Create clonality table
        clonality_table = defaultdict(list)
        for window in clonality_by_window:
            for position_pair in clonality_by_window[window]:
                clonality_table['scaffold'].append(window.split(":")[0])
                clonality_table['window_name'].append(window)
                clonality_table['position'].append(position_pair[0])
                clonality_table['clonality'].append(position_pair[1])
                clonality_table['coverage'].append(coverages[position_pair[0]])

        clonality_table_final = pd.DataFrame(clonality_table)


        # Final statistics
        logging.debug("Total SNVs-sites: " + str(total_snv_sites))
        logging.debug("Total SNV-bases: " + str(alpha_snvs))
        logging.debug("Mean clonality: " + str(float(sum(clonality_table['clonality'])) / float(len(clonality_table['clonality'])) ))
        logging.debug("Total sites above min-coverage: " + str(total_positions))
        logging.debug("Mean coverage: " + str(float(sum(coverages.values())) / len(coverages)))
        logging.debug("Total number of bases: " + str(sum(coverages.values())))

        self.coverages = coverages
        self.alpha_snvs = alpha_snvs
        self.total_snv_sites = total_snv_sites
        self.clonality_table = clonality_table_final
        self.snv_table = snv_table
        self.read_to_snvs = read_to_snvs
        self.total_positions = total_positions
        self.windows_to_snvs = windows_to_snvs
        self.all_counts = all_counts
        self.snv_counts = snv_counts

        self.results = True

def _get_base_counts(pileupcolumn, filtered_reads):
    '''
    From a pileupcolumn object, return a list with the counts of [A, C, T, G]
    '''
    P2C = {'A':0, 'C':1, 'T':2, 'G':3}
    C2P = {0:'A', 1:'C', 2:'T', 3:'G'}

    counts = [0,0,0,0]
    empty = [0,0,0,0]

    for pileupread in pileupcolumn.pileups:
        # logging.info(pileupread.)

        if not pileupread.is_del and not pileupread.is_refskip:
            read_name = pileupread.alignment.query_name
            if read_name in filtered_reads:
                # if pileupread.indel != 0:
                #     logging.info(pileupread.alignment.get_aligned_pairs())
                try:
                    counts[P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
                except KeyError: # This would be like an N or something not A/C/T/G
                    pass
    if counts != empty:
        return counts
    else:
        return False

def call_snv_site_old(counts, null_model, min_cov=5, min_freq=0.05):
    '''
    Determines whether a site has a variant based on its nucleotide count frequencies.
    '''
    P2C = {'A':0, 'C':1, 'T':2, 'G':3}
    C2P = {0:'A', 1:'C', 2:'T', 3:'G'}
    total =  sum(counts)

    if total >= min_cov:
        i = 0
        for c in counts:
            if c >= null_model[total] and float(c) / total >= min_freq:
                i += 1
        if i > 1:
            return C2P[counts.index(max(counts))]
    else:
        return False

def major_allele(snv, counts):
    d = {'A': counts[0], 'C': counts[1], 'T': counts[2], 'G': counts[3] }
    nucl = sorted(d, key=d.get, reverse=True)[0]
    return [snv + ":" + nucl, d[nucl]]

def major_allele2(counts):
    d = {'A': counts[0], 'C': counts[1], 'T': counts[2], 'G': counts[3] }
    return sorted(d, key=d.get, reverse=True)[0]

def minor_allele(snv, counts):
    d = {'A': counts[0], 'C': counts[1], 'T': counts[2], 'G': counts[3] }
    nucl = sorted(d, key=d.get, reverse=True)[1]
    return [snv + ":" + nucl, d[nucl]]

def minor_allele2(counts):
    d = {'A': counts[0], 'C': counts[1], 'T': counts[2], 'G': counts[3] }
    return sorted(d, key=d.get, reverse=True)[1]


'''
           __..--''``---....___   _..._    __
 /// //_.-'    .-/";  `        ``<._  ``.''_ `. / // /
///_.-' _..--.'_    \                    `( ) ) // //
/ (_..-' // (< _     ;_..__               ; `' / ///
 / // // //  `-._,_)' // / ``--...____..-' /// / //

 Below this is super deprecated (used nowhere)
'''

#!/usr/bin/env python

'''
Possible names for this program:
* strainRep - not a legit acronym
* blockStrain - not a legit acronym
* destrain - not a legit acronym
* instrain - not a legit acronym
* metaIPD - metagenome interference of population structure - a legt acronym
'''

# Get the version
from ._version import __version__

# Import
import os
import csv
import sys
import math
import pysam
import pickle
import logging
import argparse
import itertools
import collections
import numpy as np
import pandas as pd
import networkx as nx
from tqdm import tqdm
from Bio import SeqIO
from scipy import stats
from subprocess import call
from collections import defaultdict
from sklearn.decomposition import PCA
import traceback

import concurrent.futures
from itertools import permutations

## local imports
import inStrain.filter_reads

def entropy2(counts):
    probs = []
    total = sum(counts)
    for c in counts:
        probs.append(float(c) / total)

    ent = 0
    for i in probs:
        if i != 0:
            ent -= i * math.log(i, math.e)
    return ent


def getopts(argv):
    opts = {}  # Empty dictionary to store key-value pairs.
    while argv:  # While there are arguments left to parse...
        if argv[0][0] == '-':  # Found a "-name value" pair.
            opts[argv[0]] = argv[1]  # Add key and value to the dictionary.
        argv = argv[1:]  # Reduce the argument list by copying it starting from index 1.
    return opts

def iterate_windows(Wdb, snv_table, mm):
    '''
    Iterate windows
    '''
    Sdb = snv_table[snv_table['mm'] >= mm].sort_values('mm')\
            .drop_duplicates(subset=['scaffold', 'position'], keep='last')\
            .sort_index().drop(columns=['mm'])
    Sdb = Sdb.set_index('position', drop=False)

    for i, row in Wdb.iterrows():
        window = (int(row['start']), int(row['end']))
        sdb = Sdb[(Sdb['position'] >= window[0]) & (Sdb['position'] <= window[1])]
        snvs = list(sdb['position'].tolist())
        if len(snvs) > 0:
            yield snvs, window, sdb

def _get_series(s1, mm, lms1):
    # already a series
    if type(s1) == pd.Series:
        if s1['mm'] >= mm:
            return s1
        else:
            return lms1
    else:
        try:
            return s1.loc[mm]
        except:
            return lms1

def _update_counts(counts, ts1, ts2, mm):
    for b in ['A', 'C', 'T', 'G']:
        if len(ts1) > 0:
            counts[0][P2C[b]] += ts1[b]
        if len(ts2) > 0:
            counts[1][P2C[b]] += ts2[b]

def calc_SNV_linkage_network(read_to_snvs):
    '''
    Calculates the SNV linkage network

    Arguments:
        read_to_snvs = mm -> read -> SNVs

    saves it as a dictionary of edges in self.snv_net.
    Writes it out to a file, genome.net for reading in through other programs like node2vec
    '''
    mm_to_snv_graph = {}
    mm_to_position_graph = {}

    #logging.info("Calculating SNV linkage networks...")
    read2snvs = {}
    for mm, tread2snvs in read_to_snvs.items():
        G=nx.Graph() # graph between SNPs  (100:A and 100:C are different nodes)
        G_pos = nx.Graph() # graph between positions; unweighted

        read2snvs = {**read2snvs, **tread2snvs}
        for read in read2snvs:
            snvs = read2snvs[read]
            for pair in itertools.combinations(snvs, 2):
                if G.has_edge(pair[0], pair[1]):
                    G[pair[0]][pair[1]]['weight'] += 1
                else:
                    G.add_edge(pair[0], pair[1], weight = 1)
                    if not G_pos.has_edge(int(pair[0].split(":")[0]), int(pair[1].split(":")[0])):
                        G_pos.add_edge(int(pair[0].split(":")[0]), int(pair[1].split(":")[0]))

        mm_to_snv_graph[mm] = G
        mm_to_position_graph[mm] = G_pos

    return mm_to_snv_graph, mm_to_position_graph

def _merge_tables_special(Cdb, Adb):
    FIX_COLS = ['ANI', 'SNPs', 'unmaskedBreadth']

    # Handle edge-case where there are no SNPs
    if len(Adb) == 0:
        for c in FIX_COLS:
            Cdb[c] = 0
        return Cdb

    # Make sure anything with a coverage has a SNP
    assert len(set(Adb['mm']) - set(Cdb['mm'])) == 0

    # Do initial merge
    Cdb = pd.merge(Cdb, Adb, on=['scaffold', 'mm'], how='outer')
    Cdb = Cdb.sort_values(['scaffold', 'mm'])
    Cdb = Cdb.reset_index(drop=True)

    # Run up NaN values. For example, in the cases
    Fdb = run_up_NaN(Cdb, FIX_COLS, on='scaffold')

    # Might as well adjust some datatypes
    pass

    return Fdb

def run_up_NaN(odb, cols, on='scaffold'):
    '''
    Take NaN values and fill them with the column above.
    For example, if you have mm of [0,1,2,3], and breadth of [.9, .92, NaN, .94],
    change the breadth to [.9, .92, .92, .94]
    If you have [0,1,2,3] and [NaN, NaN, .92, .94], it'll change it to:
    [0, 0, .92, .94]
    NOTE: Must be sorted / indexed in the order that you want run up
    Args:
        odb: original dataframe
        cols: columns to do this filling on
        on: the "key" of the dataframe. If this is all NaN, fill with 0s
    Returns:
        DataFrame: new dataframe
    '''
    Fdb = odb.copy()
    for scaff, db in odb.groupby(on):
        # handle edge case where all are NaN
        if len(db[cols[0]].dropna()) == 0:
            for i, row in db.iterrows():
                for col in cols:
                    Fdb.at[i, col] = 0
            continue

        # do the actual run-up
        top = True
        for i, row in db.iterrows():
            # hangle edge case where top values are NaN
            if top & np.isnan(row[cols[0]]):
                for col in cols:
                    Fdb.at[i, col] = 0
                continue
            else:
                top = False

            # The normal run-up case
            if np.isnan(row['ANI']):
                for col in cols:
                    Fdb.at[i, col] = Fdb.at[i-1, col]

    return Fdb

def _zeros(mLen=1):
    return np.zeros(mLen, dtype=int)

def _clonT_to_clons(clonT, maxMM):
    '''
    Return the max clonality at each position
    '''
    counts = None
    for mm, count in [(mm, count) for mm, count in clonT.items() if mm <= maxMM]:
        if counts is None:
            counts = count
        else:
            counts = np.maximum(counts, count)

    if counts is None:
        return np.full(1, fill_value=np.nan)

    else:
        return counts

def make_ANI_table(snpsCounted, basesCounted, lengt, scaff, covT,
            min_cov, Wdb):
    '''
    Fill in the SNP table with the SNPs and unmaskedBreadth for each scaffold and mm
    '''
    table = defaultdict(list)
    # fill in all SNP information
    for mm in sorted(list(covT.keys())):
        covs = _mm_counts_to_counts(covT, mm)
        if covs == [0,0,0,0]:
            counted_basesO = 0
        else:
            zeros = (covs < min_cov).sum()
            counted_basesO = lengt - zeros

        counted_snps = _calc_counted_bases(snpsCounted, mm)
        counted_bases = _calc_counted_bases(basesCounted, mm)

        table['scaffold'].append(scaff)
        table['mm'].append(mm)
        table['SNPs'].append(counted_snps)
        #table['unmaskedBreadth'].append(counted_bases / lengt)
        if counted_bases == 0:
            table['ANI'].append(0)
        else:
            table['ANI'].append((counted_bases - counted_snps)/ counted_bases)

    return pd.DataFrame(table)

def parse_arguments(args):
    parser = argparse.ArgumentParser(description="inStrain version {0}".format(__version__),
             formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Required positional arguments
    parser.add_argument("bam", help="Sorted .bam file")
    parser.add_argument("fasta", help="Fasta file the bam is mapped to")

    # Optional arguments
    parser.add_argument("-o", "--output", action="store", default='inStrain', \
        help='Output prefix')
    parser.add_argument("-p", "--processes", action="store", default=6, type=int, \
        help='Threads to use for multiprocessing')
    parser.add_argument("-c", "--min_cov", action="store", default=5, \
        help='Minimum SNV coverage')
    parser.add_argument("-s", "--min_snp", action="store", default=20, \
        help='Absolute minimum number of reads connecting two SNPs to calculate LD between them.')
    parser.add_argument("-f", "--min_freq", action="store", default=0.05, \
        help='Minimum SNP frequency to confirm a SNV (both this AND the 0.  001 percent FDR snp count cutoff must be true).')
    parser.add_argument("-fdr", "--fdr", action="store", default=1e-6, type=float,\
        help='SNP false discovery rate- based on simulation data with a 0.1 percent error rate (Q30)')
    parser.add_argument("--min_fasta_reads", action="store", default=0, type=int,\
        help='Minimum number of reads mapping to a scaffold to proceed with profiling it')
    parser.add_argument("--scaffolds_to_profile", action="store",\
        help='Path to a file containing a list of scaffolds to profile- if provided will ONLY profile those scaffolds')

    # Read filtering cutoffs
    parser.add_argument("-l", "--filter_cutoff", action="store", default=0.95, type=float, \
        help='Minimum percent identity of read pairs to consensus to use the reads. Must be >, not >=')
    parser.add_argument("--min_mapq", action="store", default=-1, type=int,\
        help='Minimum mapq score of EITHER read in a pair to use that pair. Must be >, not >=')
    parser.add_argument("--max_insert_relative", action="store", default=3, type=float, \
        help='Multiplier to determine maximum insert size between two reads - default is to use 3x median insert size. Must be >, not >=')
    parser.add_argument("--min_insert", action="store", default=50, type=int,\
        help='Minimum insert size between two reads - default is 50 bp. If two reads are 50bp each and overlap completely, their insert will be 50. Must be >, not >=')

    parser.add_argument('--store_everything', action='store_true', default=False,\
        help="Store intermediate dictionaries in the pickle file; will result in significantly more RAM and disk usage")
    parser.add_argument('--skip_mm_profiling', action='store_true', default=False,\
        help="Dont perform analysis on an mm level; saves RAM and time")

    # parser.add_argument("-g", "--genes", action="store", default=None, \
    #     help='Optional genes file')
    parser.add_argument('--debug', action='store_true', default=False, \
        help ="Produce some extra output helpful for debugging")

    # Parse
    if (len(args) == 0 or args[0] == '-h' or args[0] == '--help'):
        parser.print_help()
        sys.exit(0)
    else:
        return parser.parse_args(args)

if __name__ == '__main__':
    """ This is executed when run from the command line """

    pass
    # Controller().main(sys.argv[1:])
    # args = parse_arguments(sys.argv[1:])
    # main(args)



# Get the version
from ._version import __version__

# Import packages
import os
import sys
import h5py
import logging
import argparse
import pandas as pd
from Bio import SeqIO
from collections import defaultdict

# Import inStrain stuff
import inStrain.profileUtilities
import inStrain.filter_reads
import inStrain.readComparer
import inStrain.GeneProfile
import inStrain.genomeUtilities
import inStrain.plottingUtilities
import inStrain.quickProfile
import inStrain.SNVprofile

class Controller():
    '''
    Controller of the whole shebang
    '''
    def main(self, args):
        ''' Parse user options and call the correct pipeline'''
        # Call the appropriate workflow
        if args.operation == "profile":
            self.profile_operation(args)

        if args.operation == "compare":
            self.compare_operation(args)

        if args.operation == "filter_reads":
            self.filter_reads_operation(args)

        if args.operation == "profile_genes":
            self.profile_genes_operation(args)

        if args.operation == "genome_wide":
            self.genome_wide_operation(args)

        if args.operation == "quick_profile":
            self.quick_profile_operation(args)

        if args.operation == "plot":
            self.plot_operation(args)

        if args.operation == "other":
            self.other_operation(args)

    def profile_operation(self, args):
        ProfileController().main(args)

    def compare_operation(self, args):
        inStrain.readComparer.main(args)

    def filter_reads_operation(self, args):
        inStrain.filter_reads.Controller().main(args)

    def profile_genes_operation(self, args):
        inStrain.GeneProfile.Controller().main(args)

    def genome_wide_operation(self, args):
        inStrain.genomeUtilities.Controller().main(args)

    def quick_profile_operation(self, args):
        inStrain.quickProfile.main(args)

    def plot_operation(self, args):
        inStrain.plottingUtilities.main(args)

    def other_operation(self, args):
        # Check if you should convert IS profile
        if args.old_IS != None:
            inStrain.SNVprofile.convert_SNVprofile(args.old_IS)

class ProfileController():
    '''
    Main controller of the profile command
    '''

    def main(self, args):
        '''
        The main method when run on the command line
        '''

        # Parse arguments
        #args = parse_arguments(sys_args)
        if args.output == 'inStrain':
            args.output = args.fasta.split(".")[0].split("/")[-1] #default prefix is now fasta prefix -alexcc 5/8/2019

        args = self.validate_arguments(args)
        message = """\
***************************************************
    ..:: inStrain profile Step 1. Filter reads ::..
***************************************************
        """
        logging.info(message)

        global s2l # make ths global so we can access it later.

        # Parse the args
        bam = args.bam
        vargs = vars(args)
        del vargs['bam']

        # Set up .fasta file
        FAdb, s2s = self.load_fasta(args)

        # Load dictionary of paired reads
        scaffolds = list(FAdb['scaffold'].unique())
        Rdic, RR = inStrain.filter_reads.load_paired_reads2(bam, scaffolds, **vargs)
        logging.info("{0:,} read pairs remain after filtering".format(RR['filtered_pairs'].tolist()[0]))

        if RR['filtered_pairs'].tolist()[0] == 0:
            logging.error("Because no read pairs remain I'm going to crash now. Maybe this is failing because you dont have paired reads (in which case you should adjust --pairing_filter option), or maybe its failing because the mapper you used uses full fasta headers (in which case you should use the flag --use_full_fasta_header)")
            raise Exception('No paired reads detected; see above message')

        # Get scaffold to paired reads (useful for multiprocessing)
        s2p = RR.set_index('scaffold')['filtered_pairs'].to_dict()

        # Filter the .fasta file
        FAdb = self.filter_fasta(FAdb, RR, args.min_fasta_reads)
        FAdb['filtered_pairs'] = FAdb['scaffold'].map(s2p)
        FAdb.sort_values('filtered_pairs', inplace=True, ascending=False)

        if args.skip_mm_profiling:
            Rset = set(Rdic.keys())
            del Rdic
            Rdic = Rset

        if args.debug:
            for att in ['Rdic', 'RR', 'FAdb', 's2s']:
                logging.debug("RAM PROFILE: reads {0} {1:.2f} Mb".format(att,
                    sys.getsizeof(eval(att))/1e6))

        # Profile the .bam file

        vargs['s2s'] = s2s
        vargs['s2p'] = s2p

        message = """\
***************************************************
.:: inStrain profile Step 2. Profile scaffolds ::..
***************************************************
        """
        logging.info(message)


        Sprofile = inStrain.profileUtilities.profile_bam(bam, FAdb, Rdic, **vargs)

        # Add the read report
        Sprofile.store('read_report', RR, 'pandas', "Report on reads")

        if args.store_everything:
            if args.skip_mm_profiling:
                Sprofile.store('Rdic', Rdic, 'pickle', 'list of filtered read pairs')
            else:
                Sprofile.store('Rdic', Rdic, 'dictionary', 'Read pair -> mismatches')

        # Save
        logging.info('Storing output')
        self.write_output(Sprofile, args)

        # Run the rest of things
        args.IS = Sprofile.location

        # See if you can profile genes as well
        message = """\
***************************************************
  .:: inStrain profile Step 3. Profile genes ::..
***************************************************
        """
        logging.info(message)
        if args.gene_file != None:
            args.IS = Sprofile.location
            Controller().profile_genes_operation(args)
        else:
            logging.info('Nevermind! You didnt include a genes file')

        # Make things genome_wide
        message = """\
***************************************************
.:: inStrain profile Step 4. Make genome-wide ::..
***************************************************
        """
        logging.info(message)
        if not args.skip_genome_wide:
            args.IS = Sprofile.location
            Controller().genome_wide_operation(args)
        else:
            logging.info('Nevermind! You chose to skip genome_wide')

        # Generate plots
        message = """\
***************************************************
 .:: inStrain profile Step 5. Generate plots ::..
***************************************************
        """
        logging.info(message)
        if not args.skip_plot_generation:
            args.IS = Sprofile.location
            args.plots = 'a'
            Controller().plot_operation(args)
        else:
            logging.info('Nevermind! You chose to skip making plots')

        # Final message
        message = """\
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

    ..:: inStrain profile finished ::..

Output tables........ {0}
Figures.............. {1}

See documentation for output descriptions - https://instrain.readthedocs.io/en/latest/

$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
        """.format(Sprofile.get_location('output'), \
            Sprofile.get_location('figures'))
        logging.info(message)

        return Sprofile

    def load_fasta(self, args):
        '''
        Load the sequences to be profiled

        Return a table listing scaffold name, start, end
        '''
        # PROFILE ALL SCAFFOLDS IN THE .FASTA FILE
        if args.use_full_fasta_header:
            scaff2sequence = SeqIO.to_dict(SeqIO.parse(args.fasta, "fasta"), key_function=_get_description)
        else:
            scaff2sequence = SeqIO.to_dict(SeqIO.parse(args.fasta, "fasta")) # set up .fasta file

        s2l = {s:len(scaff2sequence[s]) for s in list(scaff2sequence.keys())} # Get scaffold2length
        Fdb = pd.DataFrame(list(s2l.items()), columns=['scaffold', 'end'])
        Fdb['start'] = 0

        if args.scaffolds_to_profile != None:
            print(args.scaffolds_to_profile)
            Fdb = Fdb[Fdb['scaffold'].isin(args.scaffolds_to_profile)]
            s2s = {scaff:seq for scaff, seq in scaff2sequence.items() if scaff in args.scaffolds_to_profile}

        if len(Fdb) == 0:
            logging.error("The provided scaffold list has no overlap with the provided .fasta file!")
            logging.error("Example scaffolds in list: {0}".format("\n".join(args.scaffolds_to_profile)))
            sys.exit()

        return Fdb, scaff2sequence # also return s2l - alexcc 5/9/2019: Nah, make it scaff2sequence (s2s) (M.O. 6/10/19)

    def load_paired_reads(self, args, FAdb):
        '''
        Load paired reads to be profiled

        Return a dictionary of read pair -> info
        '''
        # Load paired reads
        scaffolds = list(FAdb['scaffold'].unique())
        scaff2pair2info, scaff2total = inStrain.filter_reads.get_paired_reads_multi(args.bam, scaffolds, processes=args.processes, ret_total=True)

        # Merge into single
        pair2info = {}
        for scaff, p2i in scaff2pair2info.items():
            for p, i in p2i.items():
                pair2info[p] = i

        # Make read report
        logging.info('Making read report')
        RR = inStrain.filter_reads.makeFilterReport(scaff2pair2info, scaff2total, pair2info=pair2info, **vars(args))

        # Filter the dictionary
        logging.info('Filtering reads')
        pair2infoF = inStrain.filter_reads.filter_paired_reads_dict(pair2info,
            **vars(args))

        # Make a report on these pair reads
        #make_read_report(pair2info, pair2infoF, args)

        return pair2infoF, RR

    def validate_arguments(self, args):
        '''
        Do some parsing, start up a logger
        '''
        # Set up "base"
        out_base = args.output

        # Set up Logger
        outbase = out_base
        RCprof = inStrain.SNVprofile.SNVprofile(outbase)
        log_loc = RCprof.get_location('log') + 'log.log'
        setup_logger(log_loc)

        # Make the bam file if you need to
        args.bam = inStrain.profileUtilities.prepare_bam_fie(args)

        # Load the list of scaffolds
        args.scaffolds_to_profile = load_scaff_list(args.scaffolds_to_profile)

        # Fix the fdr
        if args.fdr == 0:
            args.fdr = 1e-6

        return args

    def write_output(self, Sprofile, args):
        '''
        Write output files
        '''
        logging.debug("Writing output files now")
        out_base = Sprofile.get_location('output') + os.path.basename(Sprofile.get('location')) + '_'

        # Write the scaffold profile
        db = Sprofile.get_nonredundant_scaffold_table()
        db.to_csv(out_base + 'scaffold_info.tsv', index=False, sep='\t')

        # Write the SNP frequencies
        db = Sprofile.get_nonredundant_snv_table()
        db.to_csv(out_base + 'SNVs.tsv', index=False, sep='\t')

        # Write the linkage
        db = Sprofile.get_nonredundant_linkage_table()
        db.to_csv(out_base + 'linkage.tsv', index=False, sep='\t')

        # Write the read report
        RR = Sprofile.get('read_report')
        inStrain.filter_reads.write_read_report(RR, out_base + 'read_report.tsv', **vars(args))

    def filter_fasta(self, FAdb, RR, min_reads=0):
        '''
        Filter the .fasta file based on the min number of mapped paired reads
        '''
        s2r = RR.set_index('scaffold')['filtered_pairs'].to_dict()
        FAdb = FAdb[[True if (s2r[s] >= min_reads) else False for s in FAdb['scaffold']]]
        return FAdb

def load_scaff_list(list):
    '''
    If this is a text file of scaffolds, return it

    If it's a .fasta file, return a list of scaffolds
    '''

    if list == None:
        return None

    # Try as it its a fasta file
    scaffs = []
    handle = open(list, "r")
    fasta = SeqIO.parse(handle, "fasta")
    for f in fasta:
        scaffs.append(f.id)

    if len(scaffs) > 0:
        handle.close()
        return(set(scaffs))

    else:
        scaffs = []
        handle.close()
        handle = open(list, "r")
        for line in handle.readlines():
            scaffs.append(line.strip())
        handle.close()
        return set(scaffs)

def setup_logger(loc):
    ''' set up logger such that DEBUG goes only to file, rest go to file and console '''

    # Cancel if a logger already exists:
    if logging.getLogger('').handlers:
        return

    # set up logging everything to file
    logging.basicConfig(level=logging.DEBUG,
                       format='%(asctime)s %(levelname)-8s %(message)s',
                       datefmt='%m-%d %H:%M',
                       filename=loc)

    # set up logging of INFO or higher to sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('%(message)s')
    console.setFormatter(formatter)

    logging.getLogger('').addHandler(console)

    logging.debug("!"*80)
    logging.debug("***Logger started up at {0}***".format(loc))
    logging.debug("Command to run inStrain was: {0}\n".format(' '.join(sys.argv)))
    logging.debug("inStrain version {0} was run \n".format(__version__))
    logging.debug("!"*80 + '\n')

def _get_description(rec):
    return rec.description

# def parse_arguments(args):
#     parser = argparse.ArgumentParser(description="inStrain version {0}".format(__version__),
#              formatter_class=argparse.ArgumentDefaultsHelpFormatter)
#
#     # Required positional arguments
#     parser.add_argument("bam", help="Sorted .bam file")
#     parser.add_argument("fasta", help="Fasta file the bam is mapped to")
#
#     # Optional arguments
#     parser.add_argument("-o", "--output", action="store", default='inStrain', \
#         help='Output prefix')
#     parser.add_argument("-p", "--processes", action="store", default=6, type=int, \
#         help='Threads to use for multiprocessing')
#     parser.add_argument("-c", "--min_cov", action="store", default=5, \
#         help='Minimum SNV coverage')
#     parser.add_argument("-s", "--min_snp", action="store", default=20, \
#         help='Absolute minimum number of reads connecting two SNPs to calculate LD between them.')
#     parser.add_argument("-f", "--min_freq", action="store", default=0.05, \
#         help='Minimum SNP frequency to confirm a SNV (both this AND the 0.  001 percent FDR snp count cutoff must be true).')
#     parser.add_argument("-fdr", "--fdr", action="store", default=1e-6, type=float,\
#         help='SNP false discovery rate- based on simulation data with a 0.1 percent error rate (Q30)')
#     parser.add_argument("--min_fasta_reads", action="store", default=0, type=int,\
#         help='Minimum number of reads mapping to a scaffold to proceed with profiling it')
#     parser.add_argument("--scaffolds_to_profile", action="store",\
#         help='Path to a file containing a list of scaffolds to profile- if provided will ONLY profile those scaffolds')
#
#     # Read filtering cutoffs
#     parser.add_argument("-l", "--filter_cutoff", action="store", default=0.95, type=float, \
#         help='Minimum percent identity of read pairs to consensus to use the reads. Must be >, not >=')
#     parser.add_argument("--min_mapq", action="store", default=-1, type=int,\
#         help='Minimum mapq score of EITHER read in a pair to use that pair. Must be >, not >=')
#     parser.add_argument("--max_insert_relative", action="store", default=3, type=float, \
#         help='Multiplier to determine maximum insert size between two reads - default is to use 3x median insert size. Must be >, not >=')
#     parser.add_argument("--min_insert", action="store", default=50, type=int,\
#         help='Minimum insert size between two reads - default is 50 bp. If two reads are 50bp each and overlap completely, their insert will be 50. Must be >, not >=')
#
#     parser.add_argument('--store_everything', action='store_true', default=False,\
#         help="Store intermediate dictionaries in the pickle file; will result in significantly more RAM and disk usage")
#     parser.add_argument('--skip_mm_profiling', action='store_true', default=False,\
#         help="Dont perform analysis on an mm level; saves RAM and time")
#
#     # parser.add_argument("-g", "--genes", action="store", default=None, \
#     #     help='Optional genes file')
#     parser.add_argument('--debug', action='store_true', default=False, \
#         help ="Produce some extra output helpful for debugging")
#
#     # Parse
#     if (len(args) == 0 or args[0] == '-h' or args[0] == '--help'):
#         parser.print_help()
#         sys.exit(0)
#     else:
#         return parser.parse_args(args)



# Import packages
import os
import sys
import time
import pysam
import psutil
import random
import logging
import resource
import traceback
import numpy as np
import pandas as pd
from tqdm import tqdm
import concurrent.futures
from subprocess import call
from concurrent import futures
from collections import defaultdict

import inStrain.linkage
import inStrain.SNVprofile
import inStrain.readComparer

from ._version import __version__

def prepare_bam_fie(args):
    '''
    Make this a .bam file
    '''
    bam = args.bam

    if bam[-4:] == '.sam':
        logging.info("You gave me a sam- I'm going to make it a .bam now")

        bam = _sam_to_bam(bam)
        bam = _sort_index_bam(bam)

    elif bam[-4:] == '.bam':
        if (os.path.exists(bam + '.bai')) | ((os.path.exists(bam[:-4] + '.bai'))):
            pass
        else:
            bam = _sort_index_bam(bam, rm_ori=False)

    if os.stat(bam).st_size == 0:
        logging.error("Failed to generated a sorted .bam file! Make sure you have "+\
            "samtools version 1.6 or greater.")
        sys.exit()

    return bam

def profile_bam(bam, Fdb, r2m, **kwargs):
    '''
    Profile the .bam file really  well.
    This is by far the meat of the program

    arguments:
        bam = location of .bam file
        Fdb = dictionary listing fasta locations to profile
        R2M = dictionary of read pair -> number of mm
    '''
    # get arguments for profiling the scaffolds
    profArgs = kwargs

    # get arguments for the wrapper
    p = int(kwargs.get('processes', 6))
    fdr = kwargs.get('fdr', 1e-6)

    global scaff2sequence
    scaff2sequence = profArgs.pop('s2s')

    # make the dictionary global
    global R2M
    R2M = r2m

    # make a global null model
    global null_model
    null_loc = os.path.dirname(__file__) + '/helper_files/NullModel.txt'
    null_model = generate_snp_model(null_loc, fdr=fdr)

    # make a list of the scaffolds to multiprocess
    scaffolds = list(Fdb['scaffold'].unique())

    # do the multiprocessing
    Sprofiles = []
    total_cmds = len([x for x in iterate_commands(Fdb, bam, profArgs)])

    if p > 1:
        Sprofiles = []
        ex = concurrent.futures.ProcessPoolExecutor(max_workers=p)

        wait_for = [ex.submit(scaffold_profile_wrapper, cmd) for cmd in iterate_commands(Fdb, bam, profArgs)]

        for f in tqdm(futures.as_completed(wait_for), total=total_cmds, desc='Profiling scaffolds'):
            try:
                results = f.result()
                for s, log in results:
                    logging.debug(log)
                    Sprofiles.append(s)
            except:
                logging.error("We had a failure profiling a scaffold! Not sure which!")

    else:
        total_cmds = len([x for x in iterate_commands(Fdb, bam, profArgs)])
        for cmd in tqdm(iterate_commands(Fdb, bam, profArgs), desc='Profiling scaffolds: ', total=total_cmds):
            results = scaffold_profile_wrapper(cmd)
            for s, log in results:
                logging.debug(log)
                Sprofiles.append(s)

    # Re-run failed scaffolds
    failed_scaffs = set(scaffolds) - set([s.scaffold for s in Sprofiles if s is not None])
    if len(failed_scaffs) > 0:
        logging.error("The following scaffolds failed scaffold profiling- I'll try again {0}".format('\n'.join(failed_scaffs)))
        fdb = Fdb[Fdb['scaffold'].isin(failed_scaffs)]
        total_cmds = len([x for x in iterate_commands(fdb, bam, profArgs)])
        for cmd in tqdm(iterate_commands(fdb, bam, profArgs), desc='Profiling scaffolds: ', total=total_cmds):
            results = scaffold_profile_wrapper(cmd)
            try:
                for s, log in results:
                    logging.debug(log)
                    Sprofiles.append(s)
            except:
                logging.error("Double failure! Here are commands: {0}".format(cmd))

    # collate results
    logging.debug("Done processing scaffolds- making SNV profile")
    SNVprof = gen_snv_profile([s for s in Sprofiles if s is not None], **kwargs)

    # return results
    return SNVprof

def iterate_Fdbs(df, chunkSize=100):
    '''
    Break up Ndbs into chunks
    '''
    numberChunks = len(df) // chunkSize + 1
    for i in range(numberChunks):
        yield (df[i*chunkSize:(i+1)*chunkSize])

def generate_snp_model(model_file, fdr=1e-6):
    '''
    The model_file contains the columns "coverage", "probability of X coverage base by random chance"
    '''
    f = open(model_file)
    model = defaultdict(int)
    for line in f.readlines():
        if 'coverage' in line:
            continue
        counts = line.split()[1:]
        coverage = line.split()[0]
        i = 0
        for count in counts:
            if float(count) < fdr:
                model[int(coverage)] = i
                break
            i += 1

    model.default_factory = lambda:max(model.values())

    return model

def scaffold_profile_wrapper(cmds):
    '''
    Take a command and profile the scaffold
    '''
    try:
        results = []
        for cmd in cmds:
            results.append(_profile_scaffold(cmd.samfile, cmd.scaffold, cmd.locations, **cmd.arguments))
        return results
    except Exception as e:
        print(e)
        traceback.print_exc()
        logging.error("whole scaffold exception- {0}".format(str(" ".join([cmd.scaffold for cmd in cmds]))))
        return pd.DataFrame({'Failed':[True]})

def iterate_commands(Fdb, bam, args):
    '''
    Make and iterate profiling commands
    Doing it in this way makes it use way less RAM
    '''
    SECONDS = 60

    processes = args.get('processes', 6)
    s2p = args.get('s2p', None)

    cmds = []
    seconds = 0
    for scaff, db in Fdb.groupby('scaffold'):
        # make this command
        cmd = profile_command()
        cmd.scaffold = scaff
        cmd.locations = db
        cmd.samfile = bam
        cmd.arguments = args

        # Add estimated seconds
        seconds += s2p[scaff]
        cmds.append(cmd)

        # See if you're done
        if seconds >= SECONDS:
            yield cmds
            seconds = 0
            cmds = []

    yield cmds

def calc_estimated_runtime(pairs):
    SLOPE_CONSTANT = 0.0061401594694834305
    return pairs * SLOPE_CONSTANT

def humanbytes(B):
   '''
   Return the given bytes as a human friendly KB, MB, GB, or TB string
   https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
   '''
   B = float(B)
   KB = float(1024)
   MB = float(KB ** 2) # 1,048,576
   GB = float(KB ** 3) # 1,073,741,824
   TB = float(KB ** 4) # 1,099,511,627,776

   if B < KB:
      return '{0} {1}'.format(B,'Bytes' if 0 == B > 1 else 'Byte')
   elif KB <= B < MB:
      return '{0:.2f} KB'.format(B/KB)
   elif MB <= B < GB:
      return '{0:.2f} MB'.format(B/MB)
   elif GB <= B < TB:
      return '{0:.2f} GB'.format(B/GB)
   elif TB <= B:
      return '{0:.2f} TB'.format(B/TB)

class profile_command():
    '''
    This class just holds all of the mumbo-jumbo needed to profile a scaffold
    '''
    def __init__(self):
        pass

def _profile_scaffold(bam, scaffold, Wdb, **kwargs):
    '''
    Profile a scaffold

    Arguments:
        bam = name of .bam file
        scaffold = name of scaffold
        Wdb = dataframe listing windows to profile
        seq = sequence of .fasta file thats mapped to
    '''
    # Log

    # if kwargs.get('debug', False):
    pid = os.getpid()
    process  = psutil.Process(os.getpid())
    bytes_used = process.memory_info().rss
    total_available_bytes = psutil.virtual_memory()
    # logging.debug("PID {0}: {4}. Starting with {1} RAM. System has {2} of {3} available".format(
    #         pid, humanbytes(bytes_used), humanbytes(total_available_bytes[1]), humanbytes(total_available_bytes[0]),
    #         scaffold))
    log_message = "\n{4} PID {0} start at {5} with {1} RAM. System has {2} of {3} available".format(
            pid, bytes_used, total_available_bytes[1], total_available_bytes[0],
            scaffold, time.time())

    # Set up a sam file
    samfile = pysam.AlignmentFile(bam)
    try:
        iter = samfile.pileup(scaffold, truncate = True, max_depth=100000,
                                stepper = 'nofilter', compute_baq= True,
                                ignore_orphans = True, ignore_overlaps = True,
                                min_base_quality = 30)
    except ValueError:
        logging.error("scaffold {0} is not in the .bam file {1}!".format(scaffold, bam))
        return None, log_message

    # Pull some arguments
    min_cov = int(kwargs.get('min_cov', 5))
    min_covR = int(kwargs.get('rarefied_coverage', 5))
    min_freq = float(kwargs.get('min_freq', .05))
    min_snp = int(kwargs.get('min_snp', 10))
    store_everything = kwargs.get('store_everything', False)
    seq = scaff2sequence[scaffold]

    # Initialize some mm dictionaries
    mLen = len(seq)
    covT = {} # Dictionary of mm -> positional coverage along the genome
    #basesCounted = {} # Dictionary of mm -> Count of bases that got through to SNP calling
    #snpsCounted = {} # Dictionary of mm -> Count of SNPs
    read_to_snvs = defaultdict(_dlist) # Dictionary of mm -> read variant -> count
    p2c = {} # dictionary of position -> cryptic SNPs
    snv2mm2counts = {} # dictionary of position to mm to counts for SNP positions

    clonT = {} # Diciontary of mm -> clonality
    clonTR = {} # Diciontary of mm -> rarefied clonality

    # Initialize some tables
    Stable = defaultdict(list) # Holds SNP information
    pileup_counts = np.zeros(shape=(mLen, 4), dtype=int) # Holds all pileup counts - alexcc 5/9/2019

    # Do per-site processing
    for pileupcolumn in iter:
        # NOTE!
        # If you have paired reads in a pileup column, only one will come back
        # starting in pysam v0.15 (which is usually the intended functionality)
        # - M.O. 3.19.19

        # Get basic base counts
        MMcounts = _get_base_counts_mm(pileupcolumn)
        _update_covT(covT, MMcounts, pileupcolumn.pos, mLen)


        # Call SNPs
        snp, bases, total_counts = _update_snp_table_T(Stable, clonT, clonTR,\
                    MMcounts, p2c,\
                    pileupcolumn.pos, scaffold, mLen, seq[pileupcolumn.pos], min_cov=min_cov, min_covR=min_covR, min_freq=min_freq)

        # add the counts for this position to the numpy array alexcc
        if store_everything:
            pileup_counts[pileupcolumn.pos,] = total_counts

        # get linked reads
        if snp:
            _update_linked_reads(read_to_snvs, pileupcolumn, MMcounts, pileupcolumn.pos, bases, scaffold=scaffold)
            snv2mm2counts[pileupcolumn.pos] = MMcounts

    # add it back
    # if len(covT) == 0:
    #     covT = np.zeros(mLen, dtype=int)

    # shrink lots of things by default
    covT = shrink_basewise(covT, 'coverage')
    clonT = shrink_basewise(clonT, 'clonality')
    clonTR = shrink_basewise(clonTR, 'clonality')
    #snpsCounted = shrink_basewise(snpsCounted, 'snpCounted')

    # Make the SNP table
    SNPTable = pd.DataFrame(Stable)

    if len(SNPTable) > 0:
        SNPTable['scaffold'] = scaffold

        # Add cryptic SNPs
        SNPTable['cryptic'] = SNPTable['position'].map(p2c)
        SNPTable['cryptic'] = SNPTable['cryptic'].fillna(False)

        # Calc base coverage
        SNPTable['baseCoverage'] = [sum([a,c,t,g]) for a,c,t,g in zip(SNPTable['A'],SNPTable['C'],SNPTable['T'],SNPTable['G'])]

    # Do per-scaffold aggregating
    CoverageTable = make_coverage_table(covT, clonT, clonTR, mLen, scaffold, Wdb, SNPTable, min_freq=min_freq)

    # Make linkage network
    mm_to_position_graph = inStrain.linkage.calc_mm_SNV_linkage_network(read_to_snvs, scaff=scaffold)

    LDdb = inStrain.linkage.calculate_ld(mm_to_position_graph, Wdb, min_snp, snv2mm2counts=snv2mm2counts,
            scaffold=scaffold)

    # Make a profile
    Sprofile = scaffold_profile(
        scaffold=scaffold,
        bam=bam,

        min_freq=min_freq,
        min_cov=min_cov,

        length=mLen,
        #pileup_counts = pileup_counts,
        windows=Wdb,

        cumulative_scaffold_table=CoverageTable,
        #raw_ANI_table=ANITable,
        raw_snp_table=SNPTable,
        raw_linkage_table=LDdb,

        mm_reads_to_snvs=read_to_snvs
        )

    # Make cummulative tables
    Sprofile.make_cumulative_tables()

    #for att in ['covT', 'snpsCounted', 'clonT']:
    for att in ['covT', 'clonT', 'clonTR']:
        setattr(Sprofile, att, eval(att))

    # store extra things if required
    if store_everything:
        for att in ['read_to_snvs', 'mm_to_position_graph', 'pileup_counts']:
            setattr(Sprofile, att, eval(att))

    # if kwargs.get('debug', False):

    pid = os.getpid()
    process  = psutil.Process(os.getpid())
    bytes_used = process.memory_info().rss
    total_available_bytes = psutil.virtual_memory()
    log_message += "\n{4} PID {0} end at {5} with {1} RAM. System has {2} of {3} available".format(
            pid, bytes_used, total_available_bytes[1], total_available_bytes[0],
            scaffold, time.time())

    return Sprofile, log_message

def _dlist():
    return defaultdict(list)

P2C = {'A':0, 'C':1, 'T':2, 'G':3}
def _get_base_counts_mm(pileupcolumn):
    '''
    From a pileupcolumn object, return a dictionary of readMismatches ->
        list with the counts of [A, C, T, G]
    '''
    table = defaultdict(_4zeros)
    for pileupread in pileupcolumn.pileups:
        if not pileupread.is_del and not pileupread.is_refskip:
            try:
                if type(R2M) == type({}):
                    table[R2M[pileupread.alignment.query_name]]\
                    [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
                elif pileupread.alignment.query_name in R2M:
                    table[0]\
                    [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
            except KeyError: # This would be like an N or something not A/C/T/G
                pass
    return table

def _4zeros():
    return np.zeros(4, dtype=int)

def shrink_basewise(scaff2mm2array, name):
    NAME2TYPE = {'coverage':'int32', 'clonality':'float32', 'snpCounted':'bool'}

    scafmm2bases = {}
    for mm in list(scaff2mm2array.keys()):
        db = pd.Series(scaff2mm2array[mm], dtype=NAME2TYPE[name]).dropna()
        db = db[db > 0]
        scafmm2bases[mm] = db
        del scaff2mm2array[mm]
    return scafmm2bases

def _update_covT(covT, MMcounts, position, mLen):
    '''
    Update covT at this position
    '''
    for mm, count in MMcounts.items():
        if mm not in covT:
            covT[mm] = np.zeros(mLen, dtype=int)
        covT[mm][position] = sum(count)

def _update_snp_table_T(Stable, clonT, clonTR, MMcounts, p2c,\
        pos, scaff, mLen, refBase, min_cov=5, min_covR=50, min_freq=.05):
    '''
    Add information to SNP table. Update basesCounted and snpsCounted

    Also update clonality

    cryptic means that it's a SNP at lower mm levels, and then not again at higher levels

    6.10.18 - v0.4.1 - Notes

    What makes it into the Stable?
    *  There must be two bases that satisfy the following conditions:
        *  There are more than min_cov number of bases counted
        *  There are more than min_freq percent of reads at the variant base
        *  The number of variant bases is more than the null model for that coverage
    *  If those are satisfied twice, the base with the highest count is stored as the "varBase",
       and the the bases with the second highest is the "refBase"

    What is anySNP?
    *  For at least one mm position, there was a SNP called

    What is bases?
    *  A set of bases that are the varBase or refBase at least once

    What are the things that are being missed?
    *  The case where theres only one base that meets the chriteria above, but its
       different from the reference base
    *  The case where >2 variant bases are present

    What is SNP table used for?
    *  Looking for specific variant positions; performing dn/ds analysis
    *  Establishing linkage patterns

    What to do about SNPs that are fixed opposite of reference?
    *  Create a "allele_count" column (formerly morphia); 2 means there are 2 bases; 1 means theres 1
        (and its opposite of reference); 3/4 means there multiple, though only
        two will be logged
        *  Right now only 2+ are being shown; so it will be easy to filter others out
    *  Throw in a check for if there's only one bases that fits that chriteria,
       is it not the reference base?
    *  Add "refbase" back into the table to make it easier to compare SNP tables

    '''
    anySNP = False
    bases = set()
    ret_counts = np.zeros(4, dtype=int)
    for mm in sorted(list(MMcounts.keys())):
        counts = _mm_counts_to_counts(MMcounts, mm)
        snp, morphia = call_snv_site(counts, refBase, min_cov=min_cov, min_freq=min_freq) # Call SNP

        # Update clonality
        if mm not in clonT:
            clonT[mm] = np.full(mLen, fill_value=np.nan, dtype="float32")
        if sum(counts) >= min_cov:
            clonT[mm][pos] = calculate_clonality(counts)

        # Update rarefied clonality
        if mm not in clonTR:
            clonTR[mm] = np.full(mLen, fill_value=np.nan, dtype="float32")
        if sum(counts) >= min_covR:
            clonTR[mm][pos] = calculate_rarefied_clonality(counts, rarefied_coverage=min_covR)

        if not snp: # means base was not counted
            continue

        elif snp != -1: # means this is a SNP

           # calculate varBase
           counts_temp = list(counts)
           counts_temp[P2C[snp]] = 0
           varbase = C2P[list(counts_temp).index(sorted(counts_temp)[-1])] # this fixes the varbase = refbase error when there's a tie - alexcc 5/8/2019

           Stable['scaffold'].append(scaff)
           Stable['position'].append(pos)
           Stable['refBase'].append(refBase)
           for b, c in zip(['A', 'C', 'T', 'G'], counts):
               Stable[b].append(c)
           Stable['conBase'].append(snp)
           Stable['varBase'].append(varbase)
           Stable['mm'].append(mm)
           Stable['allele_count'].append(morphia)

           if morphia >= 2:
               anySNP = True
               bases.add(snp)
               bases.add(varbase)
               ret_counts = counts

           elif (morphia == 1) & (anySNP == True):
               p2c[pos] = True

           # if mm not in snpsCounted:
           #     snpsCounted[mm] = np.zeros(mLen, dtype=bool)
           # snpsCounted[mm][pos] = True

        # if it's now not a SNP, but it was in the past, mark it cryptic
        elif (snp == -1) & (anySNP == True):
            p2c[pos] = True

        # if mm not in basesCounted:
        #     basesCounted[mm] = np.zeros(mLen, dtype=bool)
        # basesCounted[mm][pos] = True # count everything that's not skipped

    return anySNP, bases, ret_counts

def _mm_counts_to_counts(MMcounts, maxMM=100):
    '''
    Take mm counts and return just counts
    '''
    counts = None
    for mm, count in [(mm, count) for mm, count in MMcounts.items() if mm <= maxMM]:
        if counts is None:
            counts = count
        else:
            counts = np.add(counts, count)

    if counts is None:
        return np.zeros(4, dtype=int)

    else:
        return counts

def _mm_counts_to_counts_shrunk(MMcounts, maxMM=100, fill_zeros=False):
    '''
    Take mm counts and return just counts
    THIS IS WITH THE SHRUNK DATAFRAME

    If fill_zeros is an int, at the end make sure the array is of that lengths
    '''
    counts = pd.Series()
    for mm, count in [(mm, count) for mm, count in MMcounts.items() if mm <= maxMM]:
        counts = counts.add(count, fill_value=0)

    if fill_zeros:
        counts = counts.append(pd.Series(np.zeros(fill_zeros - len(counts))))

    return counts


P2C = {'A':0, 'C':1, 'T':2, 'G':3}
C2P = {0:'A', 1:'C', 2:'T', 3:'G'}
def call_snv_site(counts, refBase, min_cov=5, min_freq=0.05, model=None):
    '''
    Determines whether a site has a variant based on its nucleotide count frequencies.

    Returns one of the following and the "morphia" (number of bases present in the reads):
        Base if SNP
        -1 if not SNP
        None if unCounted

    A base is considered present in the reads if:
    *  There are more than min_cov number of bases counted
    *  There are more than min_freq percent of reads at the variant base
    *  The number of variant bases is more than the null model for that coverage

    The morphia is the number of bases present in the reads. Things are returned as SNPs if:
    *  Morphia >= 2
    *  Morphia == 1, and the base present is not the genome reference base
    *  Morphia == 0 (this is a special case meaning no bases are really present)
    '''
    # Get the null model
    if model: #alexcc - so you can call this function from outside of the file
        model_to_use = model
    else:
        model_to_use = null_model

    # Make sure you have the coverage
    total = sum(counts)
    if total < min_cov:
        return None, 0

    # Count how many bases are there
    i = 0
    for c in counts:
        if c >= model_to_use[total] and float(c) / total >= min_freq:
            i += 1

    # If there are 2, say that
    if i > 1:
        return C2P[np.argmax(counts)], i

    # If theres only 1, see if its the reference base
    elif (i == 1) & (C2P[np.argmax(counts)] != refBase):
        return C2P[np.argmax(counts)], i

    # That means this is a super polymorphic position with no dominant bases
    elif i == 0:
        return C2P[np.argmax(counts)], i

    # This means its just not a SNP; one dominant reference base
    else:
        return -1, i


def calculate_clonality(counts):
    '''
    Calculates the probability that two reads have the same allele at a position (per site nucleotide diversity)
    '''
    s = sum(counts)
    prob = (float(counts[0]) / s) * (float(counts[0]) / s) + (float(counts[1]) / s) * (float(counts[1]) / s) + (float(counts[2]) / s) * (float(counts[2]) / s) + (float(counts[3]) / s) * (float(counts[3]) / s)
    return prob

def calculate_rarefied_clonality(counts, rarefied_coverage=50):
    '''
    Calculates the probability that two reads have the same allele at a position (per site nucleotide diversity)
    '''
    # Figure out the probability distribution
    s = sum(counts)
    p = [counts[i]/s for i in [0,1,2,3]]

    # Make rarefied counts
    rcounts_list = np.random.choice([0,1,2,3], rarefied_coverage, p=p)
    unique, ncounts = np.unique(rcounts_list, return_counts=True)
    item2counts = dict(zip(unique, ncounts))
    rcounts = [item2counts[i] if i in item2counts else 0 for i in [0,1,2,3]]

    return calculate_clonality(rcounts)

def get_lowest_mm(clonT, mm):
    mms = [int(m) for m in list(clonT.keys()) if int(m) <= int(mm)]
    if len(mms) == 0:
        return None
    else:
        return max(mms)

def _get_basewise_clons2(clonT, MM, fill_zeros=False):
    p2c = {}
    mms = sorted([int(mm) for mm in list(clonT.keys()) if int(mm) <= int(MM)])
    for mm in mms:
        p2c.update(clonT[mm].to_dict())

    counts = list(p2c.values())

    if fill_zeros:
        counts = counts.append(pd.Series(np.zeros(fill_zeros - len(counts))))

    return counts

def _calc_snps(Odb, mm, min_freq=0.05):
    '''
    Calculate the number of reference SNPs, bi-allelic SNPs, multi-allelic SNPs (>2), total SNPs, consensus_SNPs, and poplation_SNPs
    '''
    if len(Odb) == 0:
        return [0, 0, 0, 0, 0, 0]

    db = Odb[Odb['mm'] <= mm].sort_values('mm').drop_duplicates(subset=['position'], keep='last')

    ref_snps = len(db[(db['allele_count'] == 1)])
    bi_snps = len(db[(db['allele_count'] == 2)])
    multi_snps = len(db[(db['allele_count'] > 2)])

    con_snps = len(db[(db['conBase'] != db['refBase']) & (db['allele_count'] > 0)])

    # One pool of population SNPs are those of morphia 1 where con != ref
    p1 = len(db[(db['refBase'] != db['conBase']) & (db['allele_count'] == 1)])

    # Another pool is when its biallelic but neither are the reference
    p2 = len(db[(db['refBase'] != db['conBase']) & (db['allele_count'] == 2) & (db['refBase'] != db['varBase'])])

    # Finally, and the hardest to detect, are morphia of three where
    p3 = 0
    pdb = db[(db['refBase'] != db['conBase']) & (db['allele_count'] == 3) & (db['refBase'] != db['varBase'])]
    for i, row in pdb.iterrows():
        if not inStrain.readComparer.is_present(int(row[row['refBase']]), int(row['baseCoverage']), null_model, float(min_freq)):
            p3 += 1

    pop_snps = p1 + p2 + p3

    return [ref_snps, bi_snps, multi_snps, len(db), con_snps, pop_snps]

def make_coverage_table(covT, clonT, clonTR, lengt, scaff, Wdb, SNPTable, min_freq=0.05, debug=False):
    '''
    Add information to the table
    Args:
        covT: list of coverage values
        lengt: length of scaffold
        scaff: name of scaffold
        Wdb: windows to process over (NOT YET SUPPORTED)
    '''
    table = defaultdict(list)
    #CLdb = _clonT_to_table(clonT)
    for mm in sorted(list(covT.keys())):

        covs = _mm_counts_to_counts_shrunk(covT, mm, fill_zeros=lengt)

        if len(covs) == 0:
            covs = pd.Series([0]*lengt)

        nonzeros = np.count_nonzero(covs)
        zeros = lengt - nonzeros

        # Get clonalities
        clons = _get_basewise_clons2(clonT, mm)
        Rclons = _get_basewise_clons2(clonTR, mm)

        counted_bases = len(clons)
        rarefied_bases = len(Rclons)
        ref_snps, bi_snps, multi_snps, counted_snps, con_snps, pop_snps = _calc_snps(SNPTable, mm, min_freq=min_freq)
        #counted_snps = _calc_counted_bases(snpsCounted, mm)

        assert len(covs) == lengt, [covs, lengt, mm]

        # fill in all coverage information
        table['scaffold'].append(scaff)
        table['length'].append(lengt)
        table['breadth'].append(nonzeros/lengt)
        table['coverage'].append(np.mean(covs))
        table['median_cov'].append(int(np.median(covs)))
        table['std_cov'].append(np.std(covs))
        table['bases_w_0_coverage'].append(zeros)

        if len(clons) > 0:
            mean_c = np.mean(clons)
            median_c = np.median(clons)
            table['mean_clonality'].append(mean_c)
            table['median_clonality'].append(median_c)
            table['mean_microdiversity'].append(1-mean_c)
            table['median_microdiversity'].append(1-median_c)
        else:
            table['mean_clonality'].append(np.nan)
            table['median_clonality'].append(np.nan)
            table['mean_microdiversity'].append(np.nan)
            table['median_microdiversity'].append(np.nan)

        if len(Rclons) > 0:
            mean_c = np.mean(Rclons)
            median_c = np.median(Rclons)
            table['rarefied_mean_microdiversity'].append(1-mean_c)
            table['rarefied_median_microdiversity'].append(1-median_c)
        else:
            table['rarefied_mean_microdiversity'].append(np.nan)
            table['rarefied_median_microdiversity'].append(np.nan)

        table['unmaskedBreadth'].append(len(clons) / lengt)
        table['rarefied_breadth'].append(rarefied_bases / lengt)
        table['expected_breadth'].append(estimate_breadth(table['coverage'][-1]))

        table['SNPs'].append(counted_snps)

        table['Reference_SNPs'].append(ref_snps)
        table['BiAllelic_SNPs'].append(bi_snps)
        table['MultiAllelic_SNPs'].append(multi_snps)

        table['consensus_SNPs'].append(con_snps)
        table['population_SNPs'].append(pop_snps)

        if counted_bases == 0:
            table['conANI'].append(0)
            table['popANI'].append(0)
        else:
            table['conANI'].append((counted_bases - con_snps)/ counted_bases)
            table['popANI'].append((counted_bases - pop_snps)/ counted_bases)

        table['mm'].append(mm)

    # if debug == True:
    #     covs = _mm_counts_to_counts(covT, max(list(covT.keys())))
    #     zero_pos = [i+1 for i,x in enumerate(covs) if x==0]
    #     print(zero_pos)
    #     print(len(zero_pos))

    return pd.DataFrame(table)

def estimate_breadth(coverage):
    '''
    Estimate breadth based on coverage

    Based on the function breadth = -1.000 * e^(0.883 * coverage) + 1.000
    '''
    import numpy as np
    return (-1) * np.exp(-1 * ((0.883) * coverage)) + 1

def _clonT_to_table(clonT):
    '''
    Based on the new clonT

    You want this to have "clonality, mm, and position"
    '''
    dbs = []
    for mm, d in clonT.items():
        db = d.to_frame(name='clonality')#pd.DataFrame(d, columns=['clonality'])
        db['mm'] = mm
        db = db.reset_index()
        db = db.rename(columns={'index':'position'})
        dbs.append(db)

    if len(dbs) > 0:
        return pd.concat(dbs)

    else:
        return pd.DataFrame()
    # for mm, x in clonT.items():
    #     items += [(c, i) for i, c in enumerate(x) if c == c]
    #     mms += [mm]*(len(items) - len(mms))
    # db = pd.DataFrame({'clonality':[i[0] for i in items], 'mm':mms,
    #                     'position':[i[1] for i in items]})
    # db['position'] = db['position'].astype(int)
    #return db

# def _clonT_to_table(clonT):
#     '''
#     Much more RAM efficient way of doing this than before!
#     '''
#     items = []
#     mms = []
#     for mm, x in clonT.items():
#         items += [(c, i) for i, c in enumerate(x) if c == c]
#         mms += [mm]*(len(items) - len(mms))
#     db = pd.DataFrame({'clonality':[i[0] for i in items], 'mm':mms,
#                         'position':[i[1] for i in items]})
#     db['position'] = db['position'].astype(int)
#     return db

def _update_linked_reads(read_to_snvs, pileupcolumn, MMcounts, position, bases,
                        min_freq=0.05, scaffold=False):
    '''
    Find and update linked reads

    try:
        table[R2M[pileupread.alignment.query_name]]\
        [P2C[pileupread.alignment.query_sequence[pileupread.query_position]]] += 1
    '''

    position = str(position)

    # Get reads to snvs
    for pileupread in pileupcolumn.pileups:
        read_name = pileupread.alignment.query_name
        if not pileupread.is_del and not pileupread.is_refskip and read_name in R2M:
            if type(R2M) == type({}):
                mm = R2M[read_name]
            else:
                mm = 0
            #local_min_snp = mm2min[mm]

            try:
                val = pileupread.alignment.query_sequence[pileupread.query_position]
                #val_counts = mm2counts[mm][P2C[val]]

                # search here for debug
                # if (scaffold == 'N5_271_010G1_scaffold_1') & ((position == '2130') | (position == '2139')):
                #     print('At position {0}, mm {1}'.format(position, mm))
                #     print("{0} - {1}".format(read_name, val))#, val_counts))

                # if value is not the consensus value
                if val in bases:
                #if val_counts >= local_min_snp and float(val_counts) / sum(mm2counts[mm]) >= min_freq:
                    # this read is of a variant position
                    read_to_snvs[mm][read_name].append(position + ":" + val)
            except KeyError: # This would be like an N or something not A/C/T/G
                pass

def _calc_counted_bases(basesCounted, maxMM):
    '''
    Return the number of bases counted at a particular mm level
    Returns the number of bases at this level and all levels below it
    '''
    counts = None
    for mm, count in [(mm, count) for mm, count in basesCounted.items() if mm <= maxMM]:
        if counts is None:
            counts = count
        else:
            counts = counts.add(count, fill_value=False)

    if counts is None:
        return 0

    else:
        return counts.astype('bool').sum()

def _make_snp_table(Stable):
    if Stable is not False:
        try:
            Sdb = pd.DataFrame(Stable)
            Sdb['scaffold'] = Sdb['scaffold'].astype('category')
            Sdb['conBase'] = Sdb['conBase'].astype('category')
        except KeyError:
            #logging.info("No SNPs detected!")
            Sdb = pd.DataFrame()
    else:
        Sdb = pd.DataFrame()

    return Sdb

def _parse_Sdb(sdb):
    '''
    Add some information to sdb
    '''
    if len(sdb) == 0:
        return sdb

    sdb['varFreq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['varBase'], sdb['baseCoverage'])]
    sdb['conFreq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['conBase'], sdb['baseCoverage'])]
    sdb['refFreq'] = [[a,c,t,g][['A','C','T','G'].index(v)]/s if v in ['A','C','T','G'] else np.nan for a,c,t,g,v,s in zip(\
                        sdb['A'], sdb['C'], sdb['T'], sdb['G'], sdb['refBase'], sdb['baseCoverage'])]

    return sdb

class scaffold_profile():
    def __init__(self, **kwargs):
        '''
        initialize all attributes to None
        '''

        self.ATTRIBUTES = (\
        'scaffold', # Filename of this object

        'fasta_loc',
        'scaffold2length', # Dictionary of scaffold 2 length
        'bam',
        'windows',
        'length',

        #'pileup_counts',  # a numpy 2d array of [A,C,T,G] counts for this scaffold. Each pos = scaf pos
        'raw_snp_table', # Contains raw SNP information on a mm level
        'raw_ANI_table', # Contains raw ANI information on a mm level
        'raw_coverage_table', # Contains raw coverage information on a mm level
        'raw_linkage_table',

        'cumulative_scaffold_table', # Cumulative coverage on mm level. Formerly "scaffoldTable.csv"
        'cumulative_snv_table', # Cumulative SNP on mm level. Formerly "snpLocations.pickle"

        'mm_reads_to_snvs',

        'scaff2covT',
        'scaff2basesCounted',
        )
        for att in self.ATTRIBUTES:
            setattr(self, att, kwargs.get(att, None))
        self.version = __version__

    def make_cumulative_tables(self):
        '''
        Make cumulative tables (still raw-looking)
        This is all on the scaffold level
        '''
        # if ((self.raw_coverage_table is not None)
        #         & (self.raw_ANI_table is not None)):
        #     self.cumulative_scaffold_table = \
        #         _merge_tables_special(self.raw_coverage_table, self.raw_ANI_table)

            # if self.scaffold == 'N5_271_010G1_scaffold_101':
            #     print(self.scaffold)
            #     print(self.raw_coverage_table[self.raw_coverage_table['mm'].isna()])
            #     print(self.raw_ANI_table[self.raw_ANI_table['mm'].isna()])
            #     print(self.cumulative_scaffold_table[self.cumulative_scaffold_table['mm'].isna()])
            #     sys.exit()
            #
            #     print(self.cumulative_scaffold_table[self.cumulative_scaffold_table.isna()])
            #     print(self.raw_coverage_table)
            #     print(self.raw_ANI_table)


        if (self.raw_snp_table is not None):
            self.cumulative_snv_table = _make_snp_table(self.raw_snp_table)
            self.cumulative_snv_table = _parse_Sdb(self.cumulative_snv_table)

def gen_snv_profile(Sprofiles, **kwargs):
    '''
    Take a bunch of scaffold profiles and make an SNVprofile
    '''
    location = kwargs.get('output', False)
    store_everything = kwargs.get('store_everything', False)

    # Merge things
    scaffold_list = []
    bam_list = []
    scaffold2length = {}

    wdbs = []
    #raw_cov_dbs = []
    #raw_ani_dbs = []
    raw_snp_dbs = []
    raw_link_dbs = []

    cumu_scaff_dbs = []
    cumu_snv_dbs = []

    scaffold_2_mm_2_read_2_snvs = {}

    for Sprof in Sprofiles:

        scaffold_list.append(Sprof.scaffold)
        bam_list.append(Sprof.bam)
        scaffold2length[Sprof.scaffold] = Sprof.length

        wdbs.append(Sprof.windows)
        #raw_cov_dbs.append(Sprof.raw_coverage_table)
        #raw_ani_dbs.append(Sprof.raw_ANI_table)
        raw_snp_dbs.append(Sprof.raw_snp_table)
        raw_link_dbs.append(Sprof.raw_linkage_table)

        cumu_scaff_dbs.append(Sprof.cumulative_scaffold_table)
        cumu_snv_dbs.append(Sprof.cumulative_snv_table)

        scaffold_2_mm_2_read_2_snvs[Sprof.scaffold] = Sprof.mm_reads_to_snvs

    # Make some dataframes
    raw_snp_table = pd.concat(raw_snp_dbs).reset_index(drop=True)
    if len(raw_snp_table) > 0:
        for col in ['A', 'C', 'G', 'T', 'mm', 'position']:
            raw_snp_table[col] = raw_snp_table[col].astype(int)
            raw_snp_table['scaffold'] = raw_snp_table['scaffold'].astype('category')
            raw_snp_table['conBase'] = raw_snp_table['conBase'].astype('category')

    # convert to numpy array
    #counts_table = np.array(counts_table)

    # Make the profile
    Sprofile = inStrain.SNVprofile.SNVprofile(location)

    # Add some things
    Sprofile.store('bam_loc', bam_list[0], 'value', 'Location of .bam file')
    Sprofile.store('scaffold_list', scaffold_list, 'list', '1d list of scaffolds, in same order as counts_table')
    Sprofile.store('scaffold2length', scaffold2length, 'dictionary', 'Dictionary of scaffold 2 length')
    Sprofile.store('window_table', pd.concat(wdbs).reset_index(drop=True),
                    'pandas', 'Windows profiled over (not sure if really used right now)')
    Sprofile.store('raw_linkage_table', pd.concat(raw_link_dbs).reset_index(drop=True),
                    'pandas', 'Raw table of linkage information')
    Sprofile.store('raw_snp_table', raw_snp_table,
                    'pandas', 'Contains raw SNP information on a mm level')
    Sprofile.store('cumulative_scaffold_table', pd.concat(cumu_scaff_dbs).reset_index(drop=True),
                    'pandas', 'Cumulative coverage on mm level. Formerly scaffoldTable.csv')
    Sprofile.store('cumulative_snv_table', pd.concat(cumu_snv_dbs).reset_index(drop=True),
                    'pandas', 'Cumulative SNP on mm level. Formerly snpLocations.pickle')

    Sprofile.store('scaffold_2_mm_2_read_2_snvs', scaffold_2_mm_2_read_2_snvs,
                    'pickle', 'crazy nonsense needed for linkage')

    # On the fly testing
    # if True:
    #     assert scaffold_2_mm_2_read_2_snvs == Sprofile.get('scaffold_2_mm_2_read_2_snvs')

    # Store extra things
    att2descr = {'covT':'Scaffold -> mm -> position based coverage',
                 #'snpsCounted':"Scaffold -> mm -> position based True/False on if a SNPs is there",
                 'clonT':"Scaffold -> mm -> position based clonality",
                 'clonTR':"Scaffold -> mm -> rarefied position based clonality",
                 'read_to_snvs':'?',
                 'mm_to_position_graph':'?'}
    #for att in ['covT', 'snpsCounted', 'clonT', 'read_to_snvs', 'mm_to_position_graph']:
    for att in ['covT', 'clonT', 'clonTR', 'read_to_snvs', 'mm_to_position_graph']:
        if hasattr(Sprofiles[0], att):
            thing = {S.scaffold:getattr(S, att) for S in Sprofiles}
            Sprofile.store(att, thing, 'special', att2descr[att])

    if store_everything:
        counts_table = list() #alexcc 5/9/2019
        for Sprof in Sprofiles:
            counts_table.append(Sprof.pileup_counts) #add pileup counts to list
        counts_table = np.array(counts_table)
        Sprofile.store('counts_table', counts_table, 'numpy', '1d numpy array of 2D counts tables for each scaffold')

    return Sprofile

def _sam_to_bam(sam):
    '''
    From the location of a .sam file, convert it to a bam file and retun the location
    '''
    if sam[-4:] != '.sam':
        print('Sam file needs to end in .sam')
        sys.exit()

    bam = sam[:-4] + '.bam'
    logging.info("Converting {0} to {1}".format(sam, bam))
    cmd = ['samtools', 'view','-S','-b', sam, '>', bam]
    print(' '.join(cmd))
    call(' '.join(cmd), shell=True)

    return bam

def _sort_index_bam(bam, rm_ori=False):
    '''
    From a .bam file, sort and index it. Remove original if rm_ori
    Return path of sorted and indexed bam
    '''
    if bam[-4:] != '.bam':
        logging.error('Bam file needs to end in .bam')
        sys.exit()

    logging.info("sorting {0}".format(bam))
    sorted_bam = bam[:-4] + '.sorted.bam'
    cmd = ['samtools', 'sort', bam, '-o', sorted_bam]
    print(' '.join(cmd))
    call(cmd)

    logging.info("Indexing {0}".format(sorted_bam))
    cmd = ['samtools', 'index', sorted_bam, sorted_bam + '.bai']
    print(' '.join(cmd))
    call(cmd)

    if rm_ori:
        logging.info("Deleting {0}".format(bam))
        os.remove(bam)

    return sorted_bam
